import textwrap
from enum import Enum
from typing import Annotated, Any

import strawberry
from strawberry import BasePermission, Info
from strawberry.permission import PermissionExtension
from strawberry.printer import print_schema
from strawberry.schema.config import StrawberryConfig
from strawberry.schema_directive import Location
from strawberry.types.unset import UNSET
from tests.conftest import skip_if_gql_32


def test_print_simple_directive():
    @strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
    class Sensitive:
        reason: str

    @strawberry.type
    class Query:
        first_name: str = strawberry.field(directives=[Sensitive(reason="GDPR")])

    expected_output = """
    directive @sensitive(reason: String!) on FIELD_DEFINITION

    type Query {
      firstName: String! @sensitive(reason: "GDPR")
    }
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_print_directive_with_name():
    @strawberry.schema_directive(
        name="sensitive", locations=[Location.FIELD_DEFINITION]
    )
    class SensitiveDirective:
        reason: str

    @strawberry.type
    class Query:
        first_name: str = strawberry.field(
            directives=[SensitiveDirective(reason="GDPR")]
        )

    expected_output = """
    directive @sensitive(reason: String!) on FIELD_DEFINITION

    type Query {
      firstName: String! @sensitive(reason: "GDPR")
    }
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


@skip_if_gql_32("formatting is different in gql 3.2")
def test_directive_on_types():
    @strawberry.input
    class SensitiveValue:
        key: str
        value: str

    @strawberry.schema_directive(locations=[Location.OBJECT, Location.FIELD_DEFINITION])
    class SensitiveData:
        reason: str
        meta: list[SensitiveValue] | None = UNSET

    @strawberry.schema_directive(locations=[Location.INPUT_OBJECT])
    class SensitiveInput:
        reason: str
        meta: list[SensitiveValue] | None = UNSET

    @strawberry.schema_directive(locations=[Location.INPUT_FIELD_DEFINITION])
    class RangeInput:
        min: int
        max: int

    @strawberry.input(directives=[SensitiveInput(reason="GDPR")])
    class Input:
        first_name: str
        age: int = strawberry.field(directives=[RangeInput(min=1, max=100)])

    @strawberry.type(directives=[SensitiveData(reason="GDPR")])
    class User:
        first_name: str
        age: int
        phone: str = strawberry.field(
            directives=[
                SensitiveData(
                    reason="PRIVATE",
                    meta=[
                        SensitiveValue(
                            key="can_share_field", value="phone_share_accepted"
                        )
                    ],
                )
            ]
        )
        phone_share_accepted: bool

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self, input: Input) -> User:
            return User(
                first_name=input.first_name,
                age=input.age,
                phone="+551191551234",
                phone_share_accepted=False,
            )

    expected_output = """
    directive @rangeInput(min: Int!, max: Int!) on INPUT_FIELD_DEFINITION

    directive @sensitiveData(reason: String!, meta: [SensitiveValue!]) on OBJECT | FIELD_DEFINITION

    directive @sensitiveInput(reason: String!, meta: [SensitiveValue!]) on INPUT_OBJECT

    input Input @sensitiveInput(reason: "GDPR") {
      firstName: String!
      age: Int! @rangeInput(min: 1, max: 100)
    }

    type Query {
      user(input: Input!): User!
    }

    type User @sensitiveData(reason: "GDPR") {
      firstName: String!
      age: Int!
      phone: String! @sensitiveData(reason: "PRIVATE", meta: [{ key: "can_share_field", value: "phone_share_accepted" }])
      phoneShareAccepted: Boolean!
    }

    input SensitiveValue {
      key: String!
      value: String!
    }
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_using_different_names_for_directive_field():
    @strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
    class Sensitive:
        reason: str = strawberry.directive_field(name="as")
        real_age: str
        real_age_2: str = strawberry.directive_field(name="real_age")

    @strawberry.type
    class Query:
        first_name: str = strawberry.field(
            directives=[Sensitive(reason="GDPR", real_age="1", real_age_2="2")]
        )

    expected_output = """
    directive @sensitive(as: String!, realAge: String!, real_age: String!) on FIELD_DEFINITION

    type Query {
      firstName: String! @sensitive(as: "GDPR", realAge: "1", real_age: "2")
    }
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_respects_schema_config_for_names():
    @strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
    class Sensitive:
        real_age: str

    @strawberry.type
    class Query:
        first_name: str = strawberry.field(directives=[Sensitive(real_age="42")])

    expected_output = """
    directive @Sensitive(real_age: String!) on FIELD_DEFINITION

    type Query {
      first_name: String! @Sensitive(real_age: "42")
    }
    """

    schema = strawberry.Schema(
        query=Query, config=StrawberryConfig(auto_camel_case=False)
    )

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_respects_schema_parameter_types_for_arguments_int():
    @strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
    class Sensitive:
        real_age: int

    @strawberry.type
    class Query:
        first_name: str = strawberry.field(directives=[Sensitive(real_age=42)])

    expected_output = """
    directive @Sensitive(real_age: Int!) on FIELD_DEFINITION

    type Query {
      first_name: String! @Sensitive(real_age: 42)
    }
    """

    schema = strawberry.Schema(
        query=Query, config=StrawberryConfig(auto_camel_case=False)
    )

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_respects_schema_parameter_types_for_arguments_list_of_ints():
    @strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
    class Sensitive:
        real_age: list[int]

    @strawberry.type
    class Query:
        first_name: str = strawberry.field(directives=[Sensitive(real_age=[42])])

    expected_output = """
    directive @Sensitive(real_age: [Int!]!) on FIELD_DEFINITION

    type Query {
      first_name: String! @Sensitive(real_age: [42])
    }
    """

    schema = strawberry.Schema(
        query=Query, config=StrawberryConfig(auto_camel_case=False)
    )

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_respects_schema_parameter_types_for_arguments_list_of_strings():
    @strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
    class Sensitive:
        real_age: list[str]

    @strawberry.type
    class Query:
        first_name: str = strawberry.field(directives=[Sensitive(real_age=["42"])])

    expected_output = """
    directive @Sensitive(real_age: [String!]!) on FIELD_DEFINITION

    type Query {
      first_name: String! @Sensitive(real_age: ["42"])
    }
    """

    schema = strawberry.Schema(
        query=Query, config=StrawberryConfig(auto_camel_case=False)
    )

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_prints_directive_on_schema():
    @strawberry.schema_directive(locations=[Location.SCHEMA])
    class Tag:
        name: str

    @strawberry.type
    class Query:
        first_name: str = strawberry.field(directives=[Tag(name="team-1")])

    schema = strawberry.Schema(query=Query, schema_directives=[Tag(name="team-1")])

    expected_output = """
    directive @tag(name: String!) on SCHEMA

    schema @tag(name: "team-1") {
      query: Query
    }

    type Query {
      firstName: String!
    }
    """

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_prints_multiple_directives_on_schema():
    @strawberry.schema_directive(locations=[Location.SCHEMA])
    class Tag:
        name: str

    @strawberry.type
    class Query:
        first_name: str

    schema = strawberry.Schema(
        query=Query, schema_directives=[Tag(name="team-1"), Tag(name="team-2")]
    )

    expected_output = """
    directive @tag(name: String!) on SCHEMA

    schema @tag(name: "team-1") @tag(name: "team-2") {
      query: Query
    }

    type Query {
      firstName: String!
    }
    """

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


@skip_if_gql_32("formatting is different in gql 3.2")
def test_prints_with_types():
    @strawberry.input
    class SensitiveConfiguration:
        reason: str

    @strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
    class Sensitive:
        config: SensitiveConfiguration

    @strawberry.type
    class Query:
        first_name: str = strawberry.field(
            directives=[Sensitive(config=SensitiveConfiguration(reason="example"))]
        )

    expected_output = """
    directive @sensitive(config: SensitiveConfiguration!) on FIELD_DEFINITION

    type Query {
      firstName: String! @sensitive(config: { reason: "example" })
    }

    input SensitiveConfiguration {
      reason: String!
    }
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_prints_with_scalar():
    SensitiveConfiguration = strawberry.scalar(str, name="SensitiveConfiguration")

    @strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
    class Sensitive:
        config: SensitiveConfiguration

    @strawberry.type
    class Query:
        first_name: str = strawberry.field(directives=[Sensitive(config="Some config")])

    expected_output = """
    directive @sensitive(config: SensitiveConfiguration!) on FIELD_DEFINITION

    type Query {
      firstName: String! @sensitive(config: "Some config")
    }

    scalar SensitiveConfiguration
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_prints_with_enum():
    @strawberry.enum
    class Reason(str, Enum):
        EXAMPLE = "example"

        __slots__ = ()

    @strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
    class Sensitive:
        reason: Reason

    @strawberry.type
    class Query:
        first_name: str = strawberry.field(
            directives=[Sensitive(reason=Reason.EXAMPLE)]
        )

    expected_output = """
    directive @sensitive(reason: Reason!) on FIELD_DEFINITION

    type Query {
      firstName: String! @sensitive(reason: EXAMPLE)
    }

    enum Reason {
      EXAMPLE
    }
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_does_not_print_definition():
    @strawberry.schema_directive(
        locations=[Location.FIELD_DEFINITION], print_definition=False
    )
    class Sensitive:
        reason: str

    @strawberry.type
    class Query:
        first_name: str = strawberry.field(directives=[Sensitive(reason="GDPR")])

    expected_output = """
    type Query {
      firstName: String! @sensitive(reason: "GDPR")
    }
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_print_directive_on_scalar():
    @strawberry.schema_directive(locations=[Location.SCALAR])
    class Sensitive:
        reason: str

    SensitiveString = strawberry.scalar(
        str, name="SensitiveString", directives=[Sensitive(reason="example")]
    )

    @strawberry.type
    class Query:
        first_name: SensitiveString

    expected_output = """
    directive @sensitive(reason: String!) on SCALAR

    type Query {
      firstName: SensitiveString!
    }

    scalar SensitiveString @sensitive(reason: "example")
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_print_directive_on_enum():
    @strawberry.schema_directive(locations=[Location.ENUM])
    class Sensitive:
        reason: str

    @strawberry.enum(directives=[Sensitive(reason="example")])
    class SomeEnum(str, Enum):
        EXAMPLE = "example"

        __slots__ = ()

    @strawberry.type
    class Query:
        first_name: SomeEnum

    expected_output = """
    directive @sensitive(reason: String!) on ENUM

    type Query {
      firstName: SomeEnum!
    }

    enum SomeEnum @sensitive(reason: "example") {
      EXAMPLE
    }
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_print_directive_on_enum_value():
    @strawberry.schema_directive(locations=[Location.ENUM_VALUE])
    class Sensitive:
        reason: str

    @strawberry.enum
    class SomeEnum(Enum):
        EXAMPLE = strawberry.enum_value(
            "example", directives=[Sensitive(reason="example")]
        )

    @strawberry.type
    class Query:
        first_name: SomeEnum

    expected_output = """
    directive @sensitive(reason: String!) on ENUM_VALUE

    type Query {
      firstName: SomeEnum!
    }

    enum SomeEnum {
      EXAMPLE @sensitive(reason: "example")
    }
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_dedupe_multiple_equal_directives():
    class MemberRoleRequired(BasePermission):
        message = "Keine Rechte"

        def has_permission(self, source, info: Info, **kwargs: Any) -> bool:
            return True

    @strawberry.interface
    class MyInterface:
        id: strawberry.ID

        @strawberry.field(
            extensions=[PermissionExtension(permissions=[MemberRoleRequired()])]
        )
        def hello(self, info: Info) -> str:
            return "world"

    @strawberry.type
    class MyType1(MyInterface):
        name: str

    @strawberry.type
    class MyType2(MyInterface):
        age: int

    @strawberry.type
    class Query:
        @strawberry.field
        def my_type(self, info: Info) -> MyInterface:
            return MyType1(id=strawberry.ID("1"), name="Hello")

    expected_output = """
    directive @memberRoleRequired on FIELD_DEFINITION

    interface MyInterface {
      id: ID!
      hello: String! @memberRoleRequired
    }

    type MyType1 implements MyInterface {
      id: ID!
      hello: String! @memberRoleRequired
      name: String!
    }

    type MyType2 implements MyInterface {
      id: ID!
      hello: String! @memberRoleRequired
      age: Int!
    }

    type Query {
      myType: MyInterface!
    }
    """

    schema = strawberry.Schema(Query, types=[MyType1, MyType2])

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()

    retval = schema.execute_sync("{ myType { id hello } }")
    assert retval.errors is None
    assert retval.data == {"myType": {"id": "1", "hello": "world"}}


def test_print_directive_on_union():
    @strawberry.type
    class A:
        a: int

    @strawberry.type
    class B:
        b: int

    @strawberry.schema_directive(locations=[Location.SCALAR])
    class Sensitive:
        reason: str

    MyUnion = Annotated[
        A | B,
        strawberry.union(name="MyUnion", directives=[Sensitive(reason="example")]),
    ]

    @strawberry.type
    class Query:
        example: MyUnion

    expected_output = """
    directive @sensitive(reason: String!) on SCALAR

    type A {
      a: Int!
    }

    type B {
      b: Int!
    }

    union MyUnion @sensitive(reason: "example") = A | B

    type Query {
      example: MyUnion!
    }
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_print_directive_on_argument():
    @strawberry.schema_directive(locations=[Location.ARGUMENT_DEFINITION])
    class Sensitive:
        reason: str

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(
            self,
            name: Annotated[
                str, strawberry.argument(directives=[Sensitive(reason="example")])
            ],
            age: Annotated[
                str, strawberry.argument(directives=[Sensitive(reason="example")])
            ],
        ) -> str:
            return f"Hello {name} of {age}"

    expected_output = """
    directive @sensitive(reason: String!) on ARGUMENT_DEFINITION

    type Query {
      hello(name: String! @sensitive(reason: "example"), age: String! @sensitive(reason: "example")): String!
    }
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


def test_print_directive_on_argument_with_description():
    @strawberry.schema_directive(locations=[Location.ARGUMENT_DEFINITION])
    class Sensitive:
        reason: str

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(
            self,
            name: Annotated[
                str,
                strawberry.argument(
                    description="Name", directives=[Sensitive(reason="example")]
                ),
            ],
            age: Annotated[
                str, strawberry.argument(directives=[Sensitive(reason="example")])
            ],
        ) -> str:
            return f"Hello {name} of {age}"

    expected_output = """
    directive @sensitive(reason: String!) on ARGUMENT_DEFINITION

    type Query {
      hello(
        \"\"\"Name\"\"\"
        name: String! @sensitive(reason: "example")
        age: String! @sensitive(reason: "example")
      ): String!
    }
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


@skip_if_gql_32("formatting is different in gql 3.2")
def test_print_directive_with_unset_value():
    @strawberry.input
    class FooInput:
        a: str | None = strawberry.UNSET
        b: str | None = strawberry.UNSET

    @strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
    class FooDirective:
        input: FooInput
        optional_input: FooInput | None = strawberry.UNSET

    @strawberry.type
    class Query:
        @strawberry.field(directives=[FooDirective(input=FooInput(a="something"))])
        def foo(self, info: strawberry.Info) -> str: ...

    schema = strawberry.Schema(query=Query)

    expected_output = """
    directive @fooDirective(input: FooInput!, optionalInput: FooInput) on FIELD_DEFINITION

    type Query {
      foo: String! @fooDirective(input: { a: "something" })
    }

    input FooInput {
      a: String
      b: String
    }
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()


@skip_if_gql_32("formatting is different in gql 3.2")
def test_print_directive_with_snake_case_arguments():
    @strawberry.input
    class FooInput:
        hello: str
        hello_world: str

    @strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
    class FooDirective:
        input: FooInput
        optional_input: FooInput | None = strawberry.UNSET

    @strawberry.type
    class Query:
        @strawberry.field(
            directives=[
                FooDirective(input=FooInput(hello="hello", hello_world="hello world"))
            ]
        )
        def foo(self, info: strawberry.Info) -> str: ...

    schema = strawberry.Schema(query=Query)

    expected_output = """
    directive @fooDirective(input: FooInput!, optionalInput: FooInput) on FIELD_DEFINITION

    type Query {
      foo: String! @fooDirective(input: { hello: "hello", helloWorld: "hello world" })
    }

    input FooInput {
      hello: String!
      helloWorld: String!
    }
    """

    schema = strawberry.Schema(query=Query)

    assert print_schema(schema) == textwrap.dedent(expected_output).strip()



from __future__ import annotations

import textwrap

import pydantic

import strawberry


def test_auto_fields():
    global User

    class UserModel(pydantic.BaseModel):
        age: int
        password: str | None
        other: float

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age: strawberry.auto
        password: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> User:
            return User(age=1, password="ABC")

    schema = strawberry.Schema(query=Query)

    expected_schema = """
    type Query {
      user: User!
    }

    type User {
      age: Int!
      password: String
    }
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()

    query = "{ user { age } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["age"] == 1



from collections.abc import AsyncIterator

import strawberry
from strawberry.directive import DirectiveLocation, DirectiveValue


@strawberry.type
class Item:
    name: str
    index: int


@strawberry.type
class Person:
    name: str
    age: int
    description: str
    address: str

    prop_a: str
    prop_b: str
    prop_c: str
    prop_d: str
    prop_e: str
    prop_f: str
    prop_g: str
    prop_h: str
    prop_i: str
    prop_j: str


def create_people(n: int):
    for i in range(n):
        yield Person(
            name=f"Person {i}",
            age=i,
            description=f"Description {i}",
            address=f"Address {i}",
            prop_a=f"Prop A {i}",
            prop_b=f"Prop B {i}",
            prop_c=f"Prop C {i}",
            prop_d=f"Prop D {i}",
            prop_e=f"Prop E {i}",
            prop_f=f"Prop F {i}",
            prop_g=f"Prop G {i}",
            prop_h=f"Prop H {i}",
            prop_i=f"Prop I {i}",
            prop_j=f"Prop J {i}",
        )


people = list(create_people(n=1_000))


@strawberry.type
class Query:
    @strawberry.field
    def hello(self) -> str:
        return "Hello World!"

    @strawberry.field
    def people(self, limit: int = 100) -> list[Person]:
        return people[:limit] if limit else people

    @strawberry.field
    def items(self, count: int) -> list[Item]:
        return [Item(name="Item", index=i) for i in range(count)]


@strawberry.type
class Subscription:
    @strawberry.subscription
    async def something(self) -> AsyncIterator[str]:
        yield "Hello World!"

    @strawberry.subscription
    async def long_running(self, count: int) -> AsyncIterator[int]:
        for i in range(count):
            yield i


@strawberry.directive(locations=[DirectiveLocation.FIELD])
def uppercase(value: DirectiveValue[str]) -> str:
    return value.upper()


schema = strawberry.Schema(query=Query, subscription=Subscription)
schema_with_directives = strawberry.Schema(
    query=Query, directives=[uppercase], subscription=Subscription
)



from libcst.codemod import CodemodTest

from strawberry.codemods.maybe_optional import ConvertMaybeToOptional


class TestConvertMaybeToOptional(CodemodTest):
    TRANSFORM = ConvertMaybeToOptional

    def test_simple_maybe(self) -> None:
        before = """
            from strawberry import Maybe

            field: Maybe[str]
        """

        after = """
            from strawberry import Maybe
            from typing import Union

            field: Maybe[Union[str, None]]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_simple_maybe_union_syntax(self) -> None:
        before = """
            from strawberry import Maybe

            field: Maybe[str]
        """

        after = """
            from strawberry import Maybe
            from typing import Union

            field: Maybe[Union[str, None]]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_strawberry_maybe(self) -> None:
        before = """
            import strawberry

            field: strawberry.Maybe[int]
        """

        after = """
            import strawberry
            from typing import Union

            field: strawberry.Maybe[Union[int, None]]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_nested_type(self) -> None:
        before = """
            from strawberry import Maybe

            field: Maybe[List[str]]
        """

        after = """
            from strawberry import Maybe
            from typing import Union

            field: Maybe[Union[List[str], None]]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_already_has_none_pipe(self) -> None:
        before = """
            from strawberry import Maybe

            field: Maybe[Union[str, None]]
        """

        after = """
            from strawberry import Maybe

            field: Maybe[Union[str, None]]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_already_has_none_union(self) -> None:
        before = """
            from strawberry import Maybe
            from typing import Union

            field: Maybe[Union[str, None]]
        """

        after = """
            from strawberry import Maybe
            from typing import Union

            field: Maybe[Union[str, None]]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_multiple_maybe_fields(self) -> None:
        before = """
            from strawberry import Maybe

            @strawberry.type
            class User:
                name: Maybe[str]
                age: Maybe[int]
                email: Maybe[str]
        """

        after = """
            from strawberry import Maybe
            from typing import Union

            @strawberry.type
            class User:
                name: Maybe[Union[str, None]]
                age: Maybe[Union[int, None]]
                email: Maybe[Union[str, None]]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_function_annotation(self) -> None:
        before = """
            from strawberry import Maybe

            def get_user() -> Maybe[User]:
                return None
        """

        after = """
            from strawberry import Maybe
            from typing import Union

            def get_user() -> Maybe[Union[User, None]]:
                return None
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_generic_type(self) -> None:
        before = """
            from strawberry import Maybe

            field: Maybe[Dict[str, Any]]
        """

        after = """
            from strawberry import Maybe
            from typing import Union

            field: Maybe[Union[Dict[str, Any], None]]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_union_type_inside_maybe(self) -> None:
        before = """
            from strawberry import Maybe
            from typing import Union

            field: Maybe[Union[str, int]]
        """

        after = """
            from strawberry import Maybe
            from typing import Union

            field: Maybe[Union[Union[str, int], None]]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)



from inline_snapshot import snapshot

from .utils.marks import requires_mypy, requires_pyright, requires_ty, skip_on_windows
from .utils.typecheck import Result, typecheck

pytestmark = [skip_on_windows, requires_pyright, requires_mypy, requires_ty]


CODE = """
import strawberry
from strawberry.directive import DirectiveLocation

@strawberry.directive(
    locations=[DirectiveLocation.FRAGMENT_DEFINITION],
    description="description.",
)
def make_int(value: str) -> int:
    '''description.'''
    try:
        return int(value)
    except ValueError:
        return 0

reveal_type(make_int)
"""


def test():
    results = typecheck(CODE)

    assert results.pyright == snapshot(
        [
            Result(
                type="information",
                message='Type of "make_int" is "StrawberryDirective[int]"',
                line=16,
                column=13,
            )
        ]
    )
    assert results.mypy == snapshot(
        [
            Result(
                type="note",
                message='Revealed type is "strawberry.directive.StrawberryDirective[int]"',
                line=16,
                column=13,
            )
        ]
    )
    assert results.ty == snapshot(
        [
            Result(
                type="information",
                message="Revealed type: `StrawberryDirective[int]`",
                line=16,
                column=13,
            ),
        ]
    )



import json

import pytest
from asgiref.sync import sync_to_async
from pytest_mock import MockerFixture

import strawberry
from strawberry.dataloader import DataLoader

try:
    import django

    DJANGO_VERSION: tuple[int, int, int] = django.VERSION
except ImportError:
    DJANGO_VERSION = (0, 0, 0)


pytestmark = [
    pytest.mark.asyncio,
    pytest.mark.skipif(
        DJANGO_VERSION < (3, 1),
        reason="Async views are only supported in Django >= 3.1",
    ),
]


def _prepare_db():
    from .app.models import Example

    return [
        Example.objects.create(name=f"This is a demo async {index}").pk
        for index in range(5)
    ]


@pytest.mark.django
@pytest.mark.django_db
async def test_fetch_data_from_db(mocker: MockerFixture):
    from django.test.client import RequestFactory

    from strawberry.django.views import AsyncGraphQLView

    from .app.models import Example

    def _sync_batch_load(keys: list[str]):
        data = Example.objects.filter(id__in=keys)

        return list(data)

    prepare_db = sync_to_async(_prepare_db)
    batch_load = sync_to_async(_sync_batch_load)

    ids = await prepare_db()

    async def idx(keys: list[str]) -> list[Example]:
        return await batch_load(keys)

    mock_loader = mocker.Mock(side_effect=idx)

    loader = DataLoader[str, Example](load_fn=mock_loader)

    @strawberry.type
    class Query:
        hello: str = "strawberry"

        @strawberry.field
        async def get_example(self, id: strawberry.ID) -> str:
            example = await loader.load(id)

            return example.name

    schema = strawberry.Schema(query=Query)

    query = f"""{{
        a: getExample(id: "{ids[0]}")
        b: getExample(id: "{ids[1]}")
        c: getExample(id: "{ids[2]}")
        d: getExample(id: "{ids[3]}")
        e: getExample(id: "{ids[4]}")
    }}"""

    factory = RequestFactory()
    request = factory.post(
        "/graphql/", {"query": query}, content_type="application/json"
    )

    response = await AsyncGraphQLView.as_view(schema=schema)(request)
    data = json.loads(response.content.decode())

    assert not data.get("errors")
    assert data["data"] == {
        "a": "This is a demo async 0",
        "b": "This is a demo async 1",
        "c": "This is a demo async 2",
        "d": "This is a demo async 3",
        "e": "This is a demo async 4",
    }

    reset_db = sync_to_async(lambda: Example.objects.all().delete())
    await reset_db()

    mock_loader.assert_called_once_with([str(id_) for id_ in ids])



class RelayNestedNodesResultPostAuthor:
    id: str
    name: str

class RelayNestedNodesResultPost:
    id: str
    title: str
    author: RelayNestedNodesResultPostAuthor

class RelayNestedNodesResult:
    post: RelayNestedNodesResultPost



from typing import List, Optional

class OperationNameResultOptionalListOfPeople:
    name: str
    age: int

class OperationNameResult:
    optional_list_of_people: Optional[list[OperationNameResultOptionalListOfPeople]]



import strawberry
from strawberry.extensions.max_tokens import MaxTokensLimiter


@strawberry.type
class Human:
    name: str
    email: str


@strawberry.type
class Query:
    @strawberry.field
    def user(self, name: str | None = None, email: str | None = None) -> Human:
        return Human(name="Jane Doe", email="jane@example.com")

    version: str
    user1: Human
    user2: Human
    user3: Human


def test_1_more_token_than_allowed():
    query = """
    {
      matt: user(name: "matt") {
        name
        email
      }
    }
    """

    result = _execute_with_max_tokens(query, 13)

    assert len(result.errors) == 1
    assert (
        result.errors[0].message
        == "Syntax Error: Document contains more than 13 tokens. Parsing aborted."
    )


def test_no_errors_exactly_max_number_of_tokens():
    query = """
    {
      matt: user(name: "matt") {
        name
      }
    }
    """

    result = _execute_with_max_tokens(query, 13)

    assert not result.errors
    assert result.data


def _execute_with_max_tokens(query: str, max_token_count: int):
    schema = strawberry.Schema(
        Query, extensions=[MaxTokensLimiter(max_token_count=max_token_count)]
    )

    return schema.execute_sync(query)



import re
from typing import Literal

import pydantic
import pytest
from pydantic import BaseModel, ValidationError, conlist

import strawberry
from strawberry.experimental.pydantic._compat import IS_PYDANTIC_V1
from strawberry.types.base import StrawberryObjectDefinition, StrawberryOptional
from tests.experimental.pydantic.utils import needs_pydantic_v1, needs_pydantic_v2


@pytest.mark.parametrize(
    ("pydantic_type", "field_type"),
    [
        (pydantic.ConstrainedInt, int),
        (pydantic.PositiveInt, int),
        (pydantic.NegativeInt, int),
        (pydantic.StrictInt, int),
        (pydantic.StrictStr, str),
        (pydantic.ConstrainedStr, str),
        (pydantic.SecretStr, str),
        (pydantic.StrictBool, bool),
        (pydantic.ConstrainedBytes, bytes),
        (pydantic.SecretBytes, bytes),
        (pydantic.EmailStr, str),
        (pydantic.AnyUrl, str),
        (pydantic.AnyHttpUrl, str),
        (pydantic.HttpUrl, str),
        (pydantic.PostgresDsn, str),
        (pydantic.RedisDsn, str),
    ]
    if IS_PYDANTIC_V1
    else [
        (pydantic.PositiveInt, int),
        (pydantic.NegativeInt, int),
        (pydantic.StrictInt, int),
        (pydantic.StrictStr, str),
        (pydantic.SecretStr, str),
        (pydantic.StrictBool, bool),
        (pydantic.SecretBytes, bytes),
        (pydantic.EmailStr, str),
        (pydantic.AnyUrl, str),
        (pydantic.AnyHttpUrl, str),
        (pydantic.HttpUrl, str),
        (pydantic.PostgresDsn, str),
        (pydantic.RedisDsn, str),
    ],
)
def test_types(pydantic_type, field_type):
    class Model(pydantic.BaseModel):
        field: pydantic_type

    @strawberry.experimental.pydantic.type(Model)
    class Type:
        field: strawberry.auto

    definition: StrawberryObjectDefinition = Type.__strawberry_definition__
    assert definition.name == "Type"

    [field] = definition.fields

    assert field.python_name == "field"
    assert field.type is field_type


@needs_pydantic_v1
@pytest.mark.parametrize(
    ("pydantic_type", "field_type"),
    [(pydantic.NoneStr, str)] if IS_PYDANTIC_V1 else [],
)
def test_types_optional(pydantic_type, field_type):
    class Model(pydantic.BaseModel):
        field: pydantic_type

    @strawberry.experimental.pydantic.type(Model)
    class Type:
        field: strawberry.auto

    definition: StrawberryObjectDefinition = Type.__strawberry_definition__
    assert definition.name == "Type"

    [field] = definition.fields

    assert field.python_name == "field"
    assert isinstance(field.type, StrawberryOptional)
    assert field.type.of_type is field_type


@needs_pydantic_v2
def test_conint():
    class Model(pydantic.BaseModel):
        field: pydantic.conint(lt=100)

    @strawberry.experimental.pydantic.type(Model)
    class Type:
        field: strawberry.auto

    definition: StrawberryObjectDefinition = Type.__strawberry_definition__
    assert definition.name == "Type"

    [field] = definition.fields

    assert field.python_name == "field"
    assert field.type is int


@needs_pydantic_v1
def test_confloat():
    class Model(pydantic.BaseModel):
        field: pydantic.confloat(lt=100.5)

    @strawberry.experimental.pydantic.type(Model)
    class Type:
        field: strawberry.auto

    definition: StrawberryObjectDefinition = Type.__strawberry_definition__
    assert definition.name == "Type"

    [field] = definition.fields

    assert field.python_name == "field"
    assert field.type is float


@needs_pydantic_v1
def test_constr():
    class Model(pydantic.BaseModel):
        field: pydantic.constr(max_length=100)

    @strawberry.experimental.pydantic.type(Model)
    class Type:
        field: strawberry.auto

    definition: StrawberryObjectDefinition = Type.__strawberry_definition__
    assert definition.name == "Type"

    [field] = definition.fields

    assert field.python_name == "field"
    assert field.type is str


@needs_pydantic_v1
def test_constrained_list():
    class User(BaseModel):
        friends: conlist(str, min_items=1)

    @strawberry.experimental.pydantic.type(model=User, all_fields=True)
    class UserType: ...

    assert UserType.__strawberry_definition__.fields[0].name == "friends"
    assert (
        UserType.__strawberry_definition__.fields[0].type_annotation.raw_annotation
        == list[str]
    )

    data = UserType(friends=[])

    with pytest.raises(
        ValidationError,
        match=re.escape(
            "ensure this value has at least 1 items "
            "(type=value_error.list.min_items; limit_value=1)",
        ),
    ):
        # validation errors should happen when converting to pydantic
        data.to_pydantic()


@needs_pydantic_v1
def test_constrained_list_nested():
    class User(BaseModel):
        friends: conlist(conlist(int, min_items=1), min_items=1)

    @strawberry.experimental.pydantic.type(model=User, all_fields=True)
    class UserType: ...

    assert UserType.__strawberry_definition__.fields[0].name == "friends"
    assert (
        UserType.__strawberry_definition__.fields[0].type_annotation.raw_annotation
        == list[list[int]]
    )


@needs_pydantic_v1
@pytest.mark.parametrize(
    "pydantic_type",
    [
        pydantic.StrBytes,
        pydantic.NoneStrBytes,
        pydantic.PyObject,
        pydantic.FilePath,
        pydantic.DirectoryPath,
        pydantic.Json,
        pydantic.PaymentCardNumber,
        pydantic.ByteSize,
        # pydantic.JsonWrapper,
    ]
    if IS_PYDANTIC_V1
    else [],
)
def test_unsupported_types(pydantic_type):
    class Model(pydantic.BaseModel):
        field: pydantic_type

    with pytest.raises(
        strawberry.experimental.pydantic.exceptions.UnsupportedTypeError
    ):

        @strawberry.experimental.pydantic.type(Model)
        class Type:
            field: strawberry.auto


def test_literal_types():
    class Model(pydantic.BaseModel):
        field: Literal["field"]

    @strawberry.experimental.pydantic.type(Model)
    class Type:
        field: strawberry.auto

    definition: StrawberryObjectDefinition = Type.__strawberry_definition__
    assert definition.name == "Type"

    [field] = definition.fields

    assert field.python_name == "field"
    assert field.type == Literal["field"]



from textwrap import dedent

import pytest

import strawberry
from strawberry.annotation import StrawberryAnnotation
from strawberry.tools import create_type
from strawberry.types.base import get_object_definition
from strawberry.types.field import StrawberryField


def test_create_type():
    @strawberry.field
    def name() -> str:
        return "foo"

    MyType = create_type("MyType", [name], description="This is a description")
    definition = get_object_definition(MyType, strict=True)

    assert definition.name == "MyType"
    assert definition.description == "This is a description"
    assert definition.is_input is False

    assert len(definition.fields) == 1

    assert definition.fields[0].python_name == "name"
    assert definition.fields[0].graphql_name is None
    assert definition.fields[0].type is str


def test_create_type_extend_and_directives():
    @strawberry.field
    def name() -> str:
        return "foo"

    MyType = create_type(
        "MyType",
        [name],
        description="This is a description",
        extend=True,
        directives=[object()],
    )
    definition = get_object_definition(MyType, strict=True)

    assert definition.name == "MyType"
    assert definition.description == "This is a description"
    assert definition.is_input is False
    assert definition.extend is True
    assert len(list(definition.directives)) == 1

    assert len(definition.fields) == 1

    assert definition.fields[0].python_name == "name"
    assert definition.fields[0].graphql_name is None
    assert definition.fields[0].type is str


def test_create_input_type():
    name = StrawberryField(
        python_name="name", type_annotation=StrawberryAnnotation(str)
    )

    MyType = create_type(
        "MyType", [name], is_input=True, description="This is a description"
    )
    definition = get_object_definition(MyType, strict=True)

    assert definition.name == "MyType"
    assert definition.description == "This is a description"
    assert definition.is_input

    assert len(definition.fields) == 1

    assert definition.fields[0].python_name == "name"
    assert definition.fields[0].graphql_name is None
    assert definition.fields[0].type is str


def test_create_interface_type():
    name = StrawberryField(
        python_name="name", type_annotation=StrawberryAnnotation(str)
    )

    MyType = create_type(
        "MyType", [name], is_interface=True, description="This is a description"
    )
    definition = get_object_definition(MyType, strict=True)

    assert definition.name == "MyType"
    assert definition.description == "This is a description"
    assert definition.is_input is False
    assert definition.is_interface

    assert len(definition.fields) == 1

    assert definition.fields[0].python_name == "name"
    assert definition.fields[0].graphql_name is None
    assert definition.fields[0].type is str


def test_create_variable_type():
    def get_name() -> str:
        return "foo"

    name = strawberry.field(name="name", resolver=get_name)

    MyType = create_type("MyType", [name])
    definition = get_object_definition(MyType, strict=True)

    assert len(definition.fields) == 1

    assert definition.fields[0].python_name == "get_name"
    assert definition.fields[0].graphql_name == "name"
    assert definition.fields[0].type is str


def test_create_type_empty_list():
    with pytest.raises(ValueError):
        create_type("MyType", [])


def test_requires_resolver_or_name_to_create_type_field():
    name = strawberry.field()

    with pytest.raises(ValueError, match="Field doesn't have a name"):
        create_type("MyType", [name])


def test_can_create_type_field_from_resolver_only():
    def get_name() -> str:
        return "foo"  # pragma: no cover

    name = strawberry.field(resolver=get_name)

    MyType = create_type("MyType", [name])

    definition = get_object_definition(MyType, strict=True)
    assert len(definition.fields) == 1
    assert definition.fields[0].python_name == "get_name"
    assert definition.fields[0].graphql_name is None


def test_can_create_type_field_from_name_only():
    first_name = strawberry.field(name="firstName", graphql_type=str)

    MyType = create_type("MyType", [first_name])

    definition = get_object_definition(MyType, strict=True)
    assert len(definition.fields) == 1
    assert definition.fields[0].python_name == "firstName"
    assert definition.fields[0].graphql_name == "firstName"


def test_can_create_type_field_from_resolver_and_name():
    def get_first_name() -> str:
        return "foo"  # pragma: no cover

    first_name = strawberry.field(name="firstName", resolver=get_first_name)

    MyType = create_type("MyType", [first_name])

    definition = get_object_definition(MyType, strict=True)
    assert len(definition.fields) == 1
    assert definition.fields[0].python_name == "get_first_name"
    assert definition.fields[0].graphql_name == "firstName"


def test_create_type_field_invalid():
    with pytest.raises(TypeError):
        create_type("MyType", [strawberry.type()])


def test_create_mutation_type():
    @strawberry.type
    class User:
        username: str

    @strawberry.mutation
    def make_user(username: str) -> User:
        return User(username=username)

    Mutation = create_type("Mutation", [make_user])
    definition = get_object_definition(Mutation, strict=True)

    assert len(definition.fields) == 1

    assert definition.fields[0].python_name == "make_user"
    assert definition.fields[0].graphql_name is None
    assert definition.fields[0].type == User


def test_create_mutation_type_with_params():
    @strawberry.type
    class User:
        username: str

    @strawberry.mutation(name="makeNewUser", description="Make a new user")
    def make_user(username: str) -> User:
        return User(username=username)

    Mutation = create_type("Mutation", [make_user])
    definition = get_object_definition(Mutation, strict=True)

    assert len(definition.fields) == 1

    assert definition.fields[0].python_name == "make_user"
    assert definition.fields[0].graphql_name == "makeNewUser"
    assert definition.fields[0].type == User
    assert definition.fields[0].description == "Make a new user"


def test_create_schema():
    @strawberry.type
    class User:
        id: strawberry.ID

    @strawberry.field
    def get_user_by_id(id: strawberry.ID) -> User:
        return User(id=id)

    Query = create_type("Query", [get_user_by_id])

    schema = strawberry.Schema(query=Query)

    sdl = """
    type Query {
      getUserById(id: ID!): User!
    }

    type User {
      id: ID!
    }
    """

    assert dedent(sdl).strip() == str(schema)

    result = schema.execute_sync(
        """
        {
            getUserById(id: "TEST") {
                id
            }
        }
    """
    )

    assert not result.errors
    assert result.data == {"getUserById": {"id": "TEST"}}



import textwrap

import strawberry
from strawberry.printer import print_schema


def test_cyclic_import():
    from .type_a import TypeA
    from .type_b import TypeB

    @strawberry.type
    class Query:
        a: TypeA
        b: TypeB

    expected = """
    type Query {
      a: TypeA!
      b: TypeB!
    }

    type TypeA {
      listOfB: [TypeB!]
      typeB: TypeB!
    }

    type TypeB {
      typeA: TypeA!
    }
    """

    schema = strawberry.Schema(Query)

    assert print_schema(schema) == textwrap.dedent(expected).strip()



# type: ignore
import dataclasses
import re
from enum import Enum
from typing import Annotated, Optional, TypeVar

import pytest

import strawberry
from strawberry.exceptions import MultipleStrawberryFieldsError
from strawberry.types.base import get_object_definition, has_object_definition
from strawberry.types.field import StrawberryField


def test_enum():
    @strawberry.enum
    class Count(Enum):
        TWO = "two"
        FOUR = "four"

    @strawberry.type
    class Animal:
        legs: Count

    field: StrawberryField = get_object_definition(Animal).fields[0]

    # TODO: Remove reference to .__strawberry_definition__ with StrawberryEnumDefinition
    assert field.type is Count.__strawberry_definition__


def test_forward_reference():
    global FromTheFuture

    @strawberry.type
    class TimeTraveler:
        origin: "FromTheFuture"

    @strawberry.type
    class FromTheFuture:
        year: int

    field: StrawberryField = get_object_definition(TimeTraveler).fields[0]

    assert field.type is FromTheFuture

    del FromTheFuture


def test_list():
    @strawberry.type
    class Santa:
        making_a: list[str]

    field: StrawberryField = get_object_definition(Santa).fields[0]

    assert field.type == list[str]


def test_literal():
    @strawberry.type
    class Fabric:
        thread_type: str

    field: StrawberryField = get_object_definition(Fabric).fields[0]

    assert field.type is str


def test_object():
    @strawberry.type
    class Object:
        proper_noun: bool

    @strawberry.type
    class TransitiveVerb:
        subject: Object

    field: StrawberryField = get_object_definition(TransitiveVerb).fields[0]

    assert field.type is Object


def test_optional():
    @strawberry.type
    class HasChoices:
        decision: bool | None

    field: StrawberryField = get_object_definition(HasChoices).fields[0]

    assert field.type == Optional[bool]


def test_type_var():
    T = TypeVar("T")

    @strawberry.type
    class Gossip:
        spill_the: T

    field: StrawberryField = get_object_definition(Gossip).fields[0]

    assert field.type == T


def test_union():
    @strawberry.type
    class Europe:
        name: str

    @strawberry.type
    class UK:
        name: str

    EU = Annotated[Europe | UK, strawberry.union("EU")]

    @strawberry.type
    class WishfulThinking:
        desire: EU

    field: StrawberryField = get_object_definition(WishfulThinking).fields[0]

    assert field.type == EU


def test_fields_with_defaults():
    @strawberry.type
    class Country:
        name: str = "United Kingdom"
        currency_code: str

    country = Country(currency_code="GBP")
    assert country.name == "United Kingdom"
    assert country.currency_code == "GBP"

    country = Country(name="United States of America", currency_code="USD")
    assert country.name == "United States of America"
    assert country.currency_code == "USD"


def test_fields_with_defaults_inheritance():
    @strawberry.interface
    class A:
        text: str
        delay: int | None = None

    @strawberry.type
    class B(A):
        attachments: list[A] | None = None

    @strawberry.type
    class C(A):
        fields: list[B]

    c_inst = C(
        text="some text",
        fields=[B(text="more text")],
    )

    assert dataclasses.asdict(c_inst) == {
        "text": "some text",
        "delay": None,
        "fields": [
            {
                "text": "more text",
                "attachments": None,
                "delay": None,
            }
        ],
    }


def test_positional_args_not_allowed():
    @strawberry.type
    class Thing:
        name: str

    with pytest.raises(
        TypeError,
        match=re.escape("__init__() takes 1 positional argument but 2 were given"),
    ):
        Thing("something")


def test_object_preserves_annotations():
    @strawberry.type
    class Object:
        a: bool
        b: Annotated[str, "something"]
        c: bool = strawberry.field(graphql_type=int)
        d: Annotated[str, "something"] = strawberry.field(graphql_type=int)

    assert Object.__annotations__ == {
        "a": bool,
        "b": Annotated[str, "something"],
        "c": bool,
        "d": Annotated[str, "something"],
    }


def test_has_object_definition_returns_true_for_object_type():
    @strawberry.type
    class Palette:
        name: str

    assert has_object_definition(Palette) is True


def test_has_object_definition_returns_false_for_enum():
    @strawberry.enum
    class Color(Enum):
        RED = "red"
        GREEN = "green"

    assert has_object_definition(Color) is False


def test_has_object_definition_returns_true_for_interface():
    @strawberry.interface
    class Node:
        id: str

    assert has_object_definition(Node) is True


def test_has_object_definition_returns_true_for_input():
    @strawberry.input
    class CreateUserInput:
        name: str

    assert has_object_definition(CreateUserInput) is True


def test_has_object_definition_returns_false_for_scalar():
    from strawberry.scalars import JSON

    assert has_object_definition(JSON) is False


def test_annotated_field_with_description():
    @strawberry.type
    class Query:
        name: Annotated[str, strawberry.field(description="The name")]

    definition = get_object_definition(Query)
    field = definition.fields[0]

    assert field.python_name == "name"
    assert field.description == "The name"
    assert field.type is str


def test_annotated_field_with_graphql_name():
    @strawberry.type
    class Query:
        name: Annotated[str, strawberry.field(name="fullName")]

    definition = get_object_definition(Query)
    field = definition.fields[0]

    assert field.python_name == "name"
    assert field.graphql_name == "fullName"
    assert field.type is str


def test_annotated_field_with_deprecation_reason():
    @strawberry.type
    class Query:
        name: Annotated[
            str, strawberry.field(deprecation_reason="Use fullName instead")
        ]

    definition = get_object_definition(Query)
    field = definition.fields[0]

    assert field.python_name == "name"
    assert field.deprecation_reason == "Use fullName instead"
    assert field.type is str


def test_annotated_field_optional():
    @strawberry.type
    class Query:
        name: Annotated[str | None, strawberry.field(description="Optional name")]

    definition = get_object_definition(Query)
    field = definition.fields[0]

    assert field.python_name == "name"
    assert field.description == "Optional name"
    assert field.type == Optional[str]


def test_annotated_field_with_default_value():
    @strawberry.type
    class Query:
        name: Annotated[str, strawberry.field(description="The name")] = "default"

    definition = get_object_definition(Query)
    field = definition.fields[0]

    assert field.python_name == "name"
    assert field.description == "The name"
    assert field.type is str

    query = Query()
    assert query.name == "default"


def test_annotated_field_with_list_type():
    @strawberry.type
    class Query:
        names: Annotated[list[str], strawberry.field(description="List of names")]

    definition = get_object_definition(Query)
    field = definition.fields[0]

    assert field.python_name == "names"
    assert field.description == "List of names"
    assert field.type == list[str]


def test_annotated_field_with_nested_type():
    @strawberry.type
    class Person:
        name: str

    @strawberry.type
    class Query:
        person: Annotated[Person, strawberry.field(description="A person")]

    definition = get_object_definition(Query)
    field = definition.fields[0]

    assert field.python_name == "person"
    assert field.description == "A person"
    assert field.type is Person


def test_annotated_field_with_metadata():
    @strawberry.type
    class Query:
        name: Annotated[
            str, strawberry.field(description="Name", metadata={"custom": "value"})
        ]

    definition = get_object_definition(Query)
    field = definition.fields[0]

    assert field.python_name == "name"
    assert field.description == "Name"
    assert field.metadata == {"custom": "value"}


@pytest.mark.raises_strawberry_exception(
    MultipleStrawberryFieldsError,
    match=(
        "Annotation for field `name` on type `Query` "
        "cannot have multiple `strawberry.field`s"
    ),
)
def test_annotated_field_multiple_raises_error():
    @strawberry.type
    class Query:
        name: Annotated[
            str,
            strawberry.field(description="First"),
            strawberry.field(description="Second"),
        ]


@pytest.mark.raises_strawberry_exception(
    MultipleStrawberryFieldsError,
    match=(
        "Annotation for field `name` on type `Query` "
        "cannot have multiple `strawberry.field`s"
    ),
)
def test_annotated_field_and_assignment_raises_error():
    """Using strawberry.field in both Annotated and assignment should raise."""

    @strawberry.type
    class Query:
        name: Annotated[str, strawberry.field(description="From Annotated")] = (
            strawberry.field(description="From assignment")
        )


def test_annotated_field_with_other_annotations():
    """Test that Annotated with non-StrawberryField annotations still works."""

    @strawberry.type
    class Query:
        name: Annotated[str, "some metadata", 123]

    definition = get_object_definition(Query)
    field = definition.fields[0]

    assert field.python_name == "name"
    assert field.type is str


def test_annotated_field_mixed_with_regular_field():
    @strawberry.type
    class Query:
        annotated_field: Annotated[str, strawberry.field(description="Annotated")]
        regular_field: str = strawberry.field(description="Regular")
        plain_field: str

    definition = get_object_definition(Query)
    fields = {f.python_name: f for f in definition.fields}

    assert fields["annotated_field"].description == "Annotated"
    assert fields["regular_field"].description == "Regular"
    assert fields["plain_field"].description is None


def test_annotated_field_schema_generation():
    """Test that Annotated fields work correctly in schema generation."""

    @strawberry.type
    class Query:
        name: Annotated[str, strawberry.field(description="The name")]
        age: Annotated[int, strawberry.field(deprecation_reason="Use birthYear")]

    schema = strawberry.Schema(query=Query)
    schema_str = str(schema)

    assert '"The name"' in schema_str
    assert "@deprecated" in schema_str
    assert "Use birthYear" in schema_str


def test_annotated_field_with_input():
    """Test that Annotated fields work correctly with input types."""

    @strawberry.input
    class CreateUserInput:
        name: Annotated[str, strawberry.field(description="User's name")]
        email: Annotated[str, strawberry.field(description="User's email")]

    definition = get_object_definition(CreateUserInput)
    fields = {f.python_name: f for f in definition.fields}

    assert fields["name"].description == "User's name"
    assert fields["email"].description == "User's email"


def test_annotated_field_with_input_default_in_schema():
    """Test that Annotated fields with defaults show up correctly in schema."""

    @strawberry.input
    class CreateUserInput:
        name: Annotated[str, strawberry.field(description="User's name")] = "Anonymous"

    @strawberry.type
    class Query:
        @strawberry.field
        def create_user(self, input: CreateUserInput) -> str:
            return input.name

    schema = strawberry.Schema(query=Query)
    schema_str = str(schema)

    # The default value should appear in the GraphQL schema
    assert '= "Anonymous"' in schema_str
    assert "User's name" in schema_str

    # Instance should use the default
    instance = CreateUserInput()
    assert instance.name == "Anonymous"


def test_annotated_field_alias_reused_across_types():
    """Test that reusing an Annotated alias across types doesn't corrupt fields."""
    NameField = Annotated[str, strawberry.field(description="The name")]

    @strawberry.type
    class TypeA:
        name: NameField

    @strawberry.type
    class TypeB:
        name: NameField

    def_a = get_object_definition(TypeA)
    def_b = get_object_definition(TypeB)

    field_a = def_a.fields[0]
    field_b = def_b.fields[0]

    # Both fields should have correct descriptions
    assert field_a.description == "The name"
    assert field_b.description == "The name"

    # Fields should be independent instances
    assert field_a is not field_b
    assert field_a.origin is TypeA
    assert field_b.origin is TypeB



from pathlib import Path

import pytest
from pytest_snapshot.plugin import Snapshot

import strawberry
from strawberry.codegen import QueryCodegen, QueryCodegenPlugin
from strawberry.codegen.plugins.python import PythonPlugin
from strawberry.codegen.plugins.typescript import TypeScriptPlugin
from strawberry.relay import Node, NodeID

HERE = Path(__file__).parent
RELAY_QUERIES = list((HERE / "relay_queries").glob("*.graphql"))


# Create Node types with NodeID
@strawberry.type
class User(Node):
    id: NodeID[str]
    name: str
    email: str | None = None


@strawberry.type
class Post(Node):
    id: NodeID[str]
    title: str
    author: User


@strawberry.type
class Query:
    @strawberry.field
    def node(self, id: strawberry.ID) -> User:
        """Single node query with variable support."""
        return User(id="user-1", name="John Doe")  # pragma: no cover

    @strawberry.field
    def users(self) -> list[User]:
        """List of nodes."""
        return []  # pragma: no cover

    @strawberry.field
    def post(self) -> Post:
        """Nested node types."""
        return Post(  # pragma: no cover
            id="post-1", title="Hello", author=User(id="user-1", name="John Doe")
        )


@pytest.fixture
def relay_schema() -> strawberry.Schema:
    return strawberry.Schema(query=Query)


@pytest.mark.parametrize(
    ("plugin_class", "plugin_name", "extension"),
    [
        (PythonPlugin, "python", "py"),
        (TypeScriptPlugin, "typescript", "ts"),
    ],
    ids=["python", "typescript"],
)
@pytest.mark.parametrize(
    "query",
    RELAY_QUERIES,
    ids=[x.name for x in (HERE / "relay_queries").glob("*.graphql")],
)
def test_relay_codegen(
    query: Path,
    plugin_class: type[QueryCodegenPlugin],
    plugin_name: str,
    extension: str,
    snapshot: Snapshot,
    relay_schema: strawberry.Schema,
):
    generator = QueryCodegen(relay_schema, plugins=[plugin_class(query)])

    result = generator.run(query.read_text())

    code = result.to_string()

    snapshot.snapshot_dir = HERE / "snapshots" / "relay" / plugin_name
    snapshot.assert_match(code, f"{query.with_suffix('').stem}.{extension}")



from libcst.codemod import CodemodTest

from strawberry.codemods.annotated_unions import ConvertUnionToAnnotatedUnion


class TestConvertConstantCommand(CodemodTest):
    TRANSFORM = ConvertUnionToAnnotatedUnion

    def test_update_union(self) -> None:
        before = """
            AUnion = strawberry.union(name="ABC", types=(Foo, Bar))
        """

        after = """
            from typing import Annotated, Union

            AUnion = Annotated[Union[Foo, Bar], strawberry.union(name="ABC")]
        """

        self.assertCodemod(
            before, after, use_pipe_syntax=False, use_typing_extensions=False
        )

    def test_update_union_typing_extensions(self) -> None:
        before = """
            AUnion = strawberry.union(name="ABC", types=(Foo, Bar))
        """

        after = """
            from typing import Annotated, Union

            AUnion = Annotated[Union[Foo, Bar], strawberry.union(name="ABC")]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_update_union_using_import(self) -> None:
        before = """
            from strawberry import union

            AUnion = union(name="ABC", types=(Foo, Bar))
        """

        after = """
            from typing import Annotated, Union

            AUnion = Annotated[Union[Foo, Bar], strawberry.union(name="ABC")]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_noop_other_union(self) -> None:
        before = """
            from potato import union

            union("A", "B")
        """

        after = """
            from potato import union

            union("A", "B")
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_update_union_positional_name(self) -> None:
        before = """
            AUnion = strawberry.union("ABC", types=(Foo, Bar))
        """

        after = """
            from typing import Annotated, Union

            AUnion = Annotated[Union[Foo, Bar], strawberry.union(name="ABC")]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_update_swapped_kwargs(self) -> None:
        before = """
            AUnion = strawberry.union(types=(Foo, Bar), name="ABC")
        """

        after = """
            from typing import Annotated, Union

            AUnion = Annotated[Union[Foo, Bar], strawberry.union(name="ABC")]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_update_union_list(self) -> None:
        before = """
            AUnion = strawberry.union(name="ABC", types=[Foo, Bar])
        """

        after = """
            from typing import Annotated, Union

            AUnion = Annotated[Union[Foo, Bar], strawberry.union(name="ABC")]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_update_positional_arguments(self) -> None:
        before = """
            AUnion = strawberry.union("ABC", (Foo, Bar))
        """

        after = """
            from typing import Annotated, Union

            AUnion = Annotated[Union[Foo, Bar], strawberry.union(name="ABC")]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_supports_directives_and_description(self) -> None:
        before = """
            AUnion = strawberry.union(
                "ABC",
                (Foo, Bar),
                description="cool union",
                directives=[object()],
            )
        """

        after = """
            from typing import Annotated, Union

            AUnion = Annotated[Union[Foo, Bar], strawberry.union(name="ABC", description="cool union", directives=[object()])]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)

    def test_noop_with_annotated_unions(self) -> None:
        before = """
            AUnion = Annotated[Union[Foo, Bar], strawberry.union(name="ABC")]
        """

        after = """
            AUnion = Annotated[Union[Foo, Bar], strawberry.union(name="ABC")]
        """

        self.assertCodemod(before, after, use_pipe_syntax=False)



import pytest

import strawberry
from strawberry import Info, relay
from strawberry.relay import GlobalID
from strawberry.relay.exceptions import (
    NodeIDAnnotationError,
    RelayWrongAnnotationError,
    RelayWrongResolverAnnotationError,
)


@strawberry.type
class NonNodeType:
    foo: str


def test_raises_error_on_unknown_node_type_in_global_id():
    @strawberry.type
    class Query:
        @strawberry.field()
        def test(self, info: Info) -> GlobalID:
            _id = GlobalID("foo", "bar")
            _id.resolve_type(info)
            return _id

    schema = strawberry.Schema(query=Query)

    result = schema.execute_sync("""
        query TestQuery {
            test
        }
    """)
    assert len(result.errors) == 1
    assert (
        result.errors[0].message
        == "Cannot resolve. GlobalID requires a GraphQL type, received `foo`."
    )


def test_raises_error_on_non_node_type_in_global_id():
    @strawberry.type
    class Query:
        @strawberry.field()
        def test(self, info: Info) -> GlobalID:
            _id = GlobalID("NonNodeType", "bar")
            _id.resolve_type(info)
            return _id

    schema = strawberry.Schema(query=Query, types=(NonNodeType,))

    result = schema.execute_sync("""
        query TestQuery {
            test
        }
    """)
    assert len(result.errors) == 1
    assert (
        result.errors[0].message == "Cannot resolve. GlobalID requires a GraphQL Node "
        "type, received `NonNodeType`."
    )


@pytest.mark.raises_strawberry_exception(
    NodeIDAnnotationError,
    match='No field annotated with `NodeID` found in "Fruit"',
)
def test_raises_error_on_missing_node_id_annotation():
    @strawberry.type
    class Fruit(relay.Node):
        code: str

    @strawberry.type
    class Query:
        @relay.connection(relay.ListConnection[Fruit])
        def fruits(self) -> list[Fruit]: ...  # pragma: no cover

    strawberry.Schema(query=Query)


@pytest.mark.raises_strawberry_exception(
    NodeIDAnnotationError,
    match='More than one field annotated with `NodeID` found in "Fruit"',
)
def test_raises_error_on_multiple_node_id_annotation():
    @strawberry.type
    class Fruit(relay.Node):
        pk: relay.NodeID[str]
        code: relay.NodeID[str]

    @strawberry.type
    class Query:
        @relay.connection(relay.ListConnection[Fruit])
        def fruits(self) -> list[Fruit]: ...  # pragma: no cover

    strawberry.Schema(query=Query)


@pytest.mark.raises_strawberry_exception(
    RelayWrongAnnotationError,
    match=(
        'Wrong annotation used on field "fruits_conn". '
        'It should be annotated with a "Connection" subclass.'
    ),
)
def test_raises_error_on_connection_missing_annotation():
    @strawberry.type
    class Fruit(relay.Node):
        pk: relay.NodeID[str]

    @strawberry.type
    class Query:
        fruits_conn: list[Fruit] = relay.connection()

    strawberry.Schema(query=Query)


@pytest.mark.raises_strawberry_exception(
    RelayWrongAnnotationError,
    match=(
        'Wrong annotation used on field "custom_resolver". '
        'It should be annotated with a "Connection" subclass.'
    ),
)
def test_raises_error_on_connection_wrong_annotation():
    @strawberry.type
    class Fruit(relay.Node):
        pk: relay.NodeID[str]

    @strawberry.type
    class Query:
        @relay.connection(list[Fruit])  # type: ignore
        def custom_resolver(self) -> list[Fruit]: ...  # pragma: no cover

    strawberry.Schema(query=Query)


@pytest.mark.raises_strawberry_exception(
    RelayWrongResolverAnnotationError,
    match=(
        'Wrong annotation used on "custom_resolver" resolver. '
        "It should be return an iterable or async iterable object."
    ),
)
def test_raises_error_on_connection_resolver_wrong_annotation():
    @strawberry.type
    class Fruit(relay.Node):
        pk: relay.NodeID[str]

    @strawberry.type
    class Query:
        @relay.connection(relay.Connection[Fruit])  # type: ignore
        def custom_resolver(self): ...  # pragma: no cover

    strawberry.Schema(query=Query)



import pytest

from strawberry.extensions.tracing.opentelemetry import OpenTelemetryExtension


@pytest.fixture
def otel_ext():
    return OpenTelemetryExtension()


class SimpleObject:
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return f"SimpleObject({self.value})"


class ComplexObject:
    def __init__(self, simple_object, value):
        self.simple_object = simple_object
        self.value = value

    def __str__(self):
        return f"ComplexObject({self.simple_object!s}, {self.value})"


def test_convert_complex_number(otel_ext):
    value = 3 + 4j
    assert otel_ext.convert_to_allowed_types(value) == "(3+4j)"


def test_convert_range(otel_ext):
    value = range(3)
    assert otel_ext.convert_to_allowed_types(value) == "0, 1, 2"


def test_convert_bytearray(otel_ext):
    value = bytearray(b"hello world")
    assert otel_ext.convert_to_allowed_types(value) == b"hello world"


def test_convert_memoryview(otel_ext):
    value = memoryview(b"hello world")
    assert otel_ext.convert_to_allowed_types(value) == b"hello world"


def test_convert_set(otel_ext):
    value = {1, 2, 3, 4}
    converted_value = otel_ext.convert_to_allowed_types(value)
    assert set(converted_value.strip("{}").split(", ")) == {"1", "2", "3", "4"}


def test_convert_frozenset(otel_ext):
    value = frozenset([1, 2, 3, 4])
    converted_value = otel_ext.convert_to_allowed_types(value)
    assert set(converted_value.strip("{}").split(", ")) == {"1", "2", "3", "4"}


def test_convert_complex_object_with_simple_object(otel_ext):
    simple_obj = SimpleObject(42)
    complex_obj = ComplexObject(simple_obj, 99)
    assert (
        otel_ext.convert_to_allowed_types(complex_obj)
        == "ComplexObject(SimpleObject(42), 99)"
    )


def test_convert_dictionary(otel_ext):
    value = {
        "int": 1,
        "float": 3.14,
        "bool": True,
        "str": "hello",
        "list": [1, 2, 3],
        "tuple": (4, 5, 6),
        "simple_object": SimpleObject(42),
    }

    expected = (
        "{int: 1, "
        "float: 3.14, "
        "bool: True, "
        "str: hello, "
        "list: 1, 2, 3, "
        "tuple: 4, 5, 6, "
        "simple_object: SimpleObject(42)}"
    )

    assert otel_ext.convert_to_allowed_types(value) == expected


def test_convert_bool(otel_ext):
    assert otel_ext.convert_to_allowed_types(True) is True
    assert otel_ext.convert_to_allowed_types(False) is False


def test_convert_str(otel_ext):
    assert otel_ext.convert_to_allowed_types("hello") == "hello"


def test_convert_bytes(otel_ext):
    assert otel_ext.convert_to_allowed_types(b"hello") == b"hello"


def test_convert_int(otel_ext):
    assert otel_ext.convert_to_allowed_types(42) == 42


def test_convert_float(otel_ext):
    assert otel_ext.convert_to_allowed_types(3.14) == 3.14


def test_convert_simple_object(otel_ext):
    obj = SimpleObject(42)
    assert otel_ext.convert_to_allowed_types(obj) == "SimpleObject(42)"


def test_convert_list_of_basic_types(otel_ext):
    value = [1, "hello", 3.14, True, False]
    assert otel_ext.convert_to_allowed_types(value) == "1, hello, 3.14, True, False"


def test_convert_list_of_mixed_types(otel_ext):
    value = [1, "hello", 3.14, SimpleObject(42)]
    assert (
        otel_ext.convert_to_allowed_types(value) == "1, hello, 3.14, SimpleObject(42)"
    )


def test_convert_tuple_of_basic_types(otel_ext):
    value = (1, "hello", 3.14, True, False)
    assert otel_ext.convert_to_allowed_types(value) == "1, hello, 3.14, True, False"


def test_convert_tuple_of_mixed_types(otel_ext):
    value = (1, "hello", 3.14, SimpleObject(42))
    assert (
        otel_ext.convert_to_allowed_types(value) == "1, hello, 3.14, SimpleObject(42)"
    )



import strawberry
from strawberry.schema.config import StrawberryConfig


def test_camel_case_is_on_by_default():
    @strawberry.type
    class Query:
        example_field: str = "Example"

    schema = strawberry.Schema(query=Query)

    query = """
        {
            __type(name: "Query") {
                fields {
                    name
                }
            }
        }
    """

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["__type"]["fields"] == [{"name": "exampleField"}]


def test_can_set_camel_casing():
    @strawberry.type
    class Query:
        example_field: str = "Example"

    schema = strawberry.Schema(
        query=Query, config=StrawberryConfig(auto_camel_case=True)
    )

    query = """
        {
            __type(name: "Query") {
                fields {
                    name
                }
            }
        }
    """

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["__type"]["fields"] == [{"name": "exampleField"}]


def test_can_set_camel_casing_to_false():
    @strawberry.type
    class Query:
        example_field: str = "Example"

    schema = strawberry.Schema(
        query=Query, config=StrawberryConfig(auto_camel_case=False)
    )

    query = """
        {
            __type(name: "Query") {
                fields {
                    name
                }
            }
        }
    """

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["__type"]["fields"] == [{"name": "example_field"}]


def test_can_set_camel_casing_to_false_uses_name():
    @strawberry.type
    class Query:
        example_field: str = strawberry.field(name="exampleField")

    schema = strawberry.Schema(
        query=Query, config=StrawberryConfig(auto_camel_case=False)
    )

    query = """
        {
            __type(name: "Query") {
                fields {
                    name
                }
            }
        }
    """

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["__type"]["fields"] == [{"name": "exampleField"}]


def test_can_set_camel_casing_to_false_uses_name_field_decorator():
    @strawberry.type
    class Query:
        @strawberry.field(name="exampleField")
        def example_field(self) -> str:
            return "ABC"

    schema = strawberry.Schema(
        query=Query, config=StrawberryConfig(auto_camel_case=False)
    )

    query = """
        {
            __type(name: "Query") {
                fields {
                    name
                }
            }
        }
    """

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["__type"]["fields"] == [{"name": "exampleField"}]


def test_camel_case_is_on_by_default_arguments():
    @strawberry.type
    class Query:
        @strawberry.field
        def example_field(self, example_input: str) -> str:
            return example_input

    schema = strawberry.Schema(query=Query)

    query = """
        {
            __type(name: "Query") {
                fields {
                    name
                    args { name }
                }
            }
        }
    """

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["__type"]["fields"] == [
        {"args": [{"name": "exampleInput"}], "name": "exampleField"}
    ]


def test_can_turn_camel_case_off_arguments():
    @strawberry.type
    class Query:
        @strawberry.field
        def example_field(self, example_input: str) -> str:
            return example_input

    schema = strawberry.Schema(
        query=Query, config=StrawberryConfig(auto_camel_case=False)
    )

    query = """
        {
            __type(name: "Query") {
                fields {
                    name
                    args { name }
                }
            }
        }
    """

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["__type"]["fields"] == [
        {"args": [{"name": "example_input"}], "name": "example_field"}
    ]


def test_can_turn_camel_case_off_arguments_conversion_works():
    @strawberry.type
    class Query:
        @strawberry.field
        def example_field(self, example_input: str) -> str:
            return example_input

    schema = strawberry.Schema(
        query=Query, config=StrawberryConfig(auto_camel_case=False)
    )

    query = """
        {
            example_field(example_input: "Hello world")
        }
    """

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["example_field"] == "Hello world"



from enum import Enum
from typing import Optional, TypeVar

import strawberry
from strawberry.annotation import StrawberryAnnotation
from strawberry.types.field import StrawberryField
from strawberry.types.union import StrawberryUnion


def test_enum():
    @strawberry.enum
    class Egnum(Enum):
        a = "A"
        b = "B"

    annotation = StrawberryAnnotation(Egnum)
    field = StrawberryField(type_annotation=annotation)

    # TODO: Remove reference to .__strawberry_definition__ with StrawberryEnumDefinition
    assert field.type is Egnum.__strawberry_definition__


def test_forward_reference():
    global RefForward

    annotation = StrawberryAnnotation("RefForward", namespace=globals())
    field = StrawberryField(type_annotation=annotation)

    @strawberry.type
    class RefForward:
        ref: int

    assert field.type is RefForward

    del RefForward


def test_list():
    annotation = StrawberryAnnotation(list[int])
    field = StrawberryField(type_annotation=annotation)

    assert field.type == list[int]


def test_literal():
    annotation = StrawberryAnnotation(bool)
    field = StrawberryField(type_annotation=annotation)

    assert field.type is bool


def test_object():
    @strawberry.type
    class TypeyType:
        value: str

    annotation = StrawberryAnnotation(TypeyType)
    field = StrawberryField(type_annotation=annotation)

    assert field.type is TypeyType


def test_optional():
    annotation = StrawberryAnnotation(Optional[float])
    field = StrawberryField(type_annotation=annotation)

    assert field.type == Optional[float]


def test_type_var():
    T = TypeVar("T")

    annotation = StrawberryAnnotation(T)
    field = StrawberryField(type_annotation=annotation)

    assert field.type == T


def test_union():
    @strawberry.type
    class Un:
        fi: int

    @strawberry.type
    class Ion:
        eld: float

    union = StrawberryUnion(
        name="UnionName",
        type_annotations=(StrawberryAnnotation(Un), StrawberryAnnotation(Ion)),
    )
    annotation = StrawberryAnnotation(union)
    field = StrawberryField(type_annotation=annotation)

    assert field.type is union



from __future__ import annotations

from io import BytesIO
from json import dumps
from random import randint
from typing import Any, Literal

from sanic import Sanic
from sanic.request import Request as SanicRequest

from strawberry.http import GraphQLHTTPResponse
from strawberry.http.ides import GraphQL_IDE
from strawberry.http.temporal_response import TemporalResponse
from strawberry.sanic.views import GraphQLView as BaseGraphQLView
from strawberry.schema import Schema
from strawberry.types import ExecutionResult
from tests.http.context import get_context
from tests.views.schema import Query

from .base import JSON, HttpClient, Response, ResultOverrideFunction


class GraphQLView(BaseGraphQLView[object, Query]):
    result_override: ResultOverrideFunction = None

    def __init__(self, *args: Any, **kwargs: Any):
        self.result_override = kwargs.pop("result_override")
        super().__init__(*args, **kwargs)

    async def get_root_value(self, request: SanicRequest) -> Query:
        await super().get_root_value(request)  # for coverage
        return Query()

    async def get_context(
        self, request: SanicRequest, response: TemporalResponse
    ) -> object:
        context = await super().get_context(request, response)

        return get_context(context)

    async def process_result(
        self, request: SanicRequest, result: ExecutionResult
    ) -> GraphQLHTTPResponse:
        if self.result_override:
            return self.result_override(result)

        return await super().process_result(request, result)


class SanicHttpClient(HttpClient):
    def __init__(
        self,
        schema: Schema,
        graphql_ide: GraphQL_IDE | None = "graphiql",
        allow_queries_via_get: bool = True,
        result_override: ResultOverrideFunction = None,
        multipart_uploads_enabled: bool = False,
    ):
        self.app = Sanic(
            f"test_{int(randint(0, 1000))}",  # noqa: S311
        )
        view = GraphQLView.as_view(
            schema=schema,
            graphql_ide=graphql_ide,
            allow_queries_via_get=allow_queries_via_get,
            result_override=result_override,
            multipart_uploads_enabled=multipart_uploads_enabled,
        )
        self.app.add_route(view, "/graphql")

    async def _graphql_request(
        self,
        method: Literal["get", "post"],
        query: str | None = None,
        operation_name: str | None = None,
        variables: dict[str, object] | None = None,
        files: dict[str, BytesIO] | None = None,
        headers: dict[str, str] | None = None,
        extensions: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Response:
        body = self._build_body(
            query=query,
            operation_name=operation_name,
            variables=variables,
            files=files,
            method=method,
            extensions=extensions,
        )

        if body:
            if method == "get":
                kwargs["params"] = body
            elif files:
                kwargs["data"] = body
            else:
                kwargs["content"] = dumps(body)

        _request, response = await self.app.asgi_client.request(
            method,
            "/graphql",
            headers=self._get_headers(method=method, headers=headers, files=files),
            files=files,
            **kwargs,
        )

        return Response(
            status_code=response.status_code,
            data=response.content,
            headers=response.headers,
        )

    async def request(
        self,
        url: str,
        method: Literal["head", "get", "post", "patch", "put", "delete"],
        headers: dict[str, str] | None = None,
    ) -> Response:
        _request, response = await self.app.asgi_client.request(
            method,
            url,
            headers=headers,
        )

        return Response(
            status_code=response.status_code,
            data=response.content,
            headers=response.headers,
        )

    async def get(
        self,
        url: str,
        headers: dict[str, str] | None = None,
    ) -> Response:
        return await self.request(url, "get", headers=headers)

    async def post(
        self,
        url: str,
        data: bytes | None = None,
        json: JSON | None = None,
        headers: dict[str, str] | None = None,
    ) -> Response:
        body = dumps(json) if json is not None else data

        _request, response = await self.app.asgi_client.request(
            "post", url, content=body, headers=headers
        )

        return Response(
            status_code=response.status_code,
            data=response.content,
            headers=response.headers,
        )



from inline_snapshot import snapshot

from .utils.marks import requires_mypy, requires_pyright, requires_ty, skip_on_windows
from .utils.typecheck import Result, typecheck

pytestmark = [skip_on_windows, requires_pyright, requires_mypy, requires_ty]


CODE = """
import strawberry


@strawberry.input
class User:
    name: str


User(name="Patrick")
User(n="Patrick")

reveal_type(User)
reveal_type(User.__init__)
"""


def test():
    results = typecheck(CODE)

    assert results.pyright == snapshot(
        [
            Result(
                type="error",
                message='Argument missing for parameter "name"',
                line=11,
                column=1,
            ),
            Result(type="error", message='No parameter named "n"', line=11, column=6),
            Result(
                type="information",
                message='Type of "User" is "type[User]"',
                line=13,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "User.__init__" is "(self: User, *, name: str) -> None"',
                line=14,
                column=13,
            ),
        ]
    )
    assert results.mypy == snapshot(
        [
            Result(
                type="error",
                message='Unexpected keyword argument "n" for "User"',
                line=11,
                column=1,
            ),
            Result(
                type="note",
                message='Revealed type is "def (*, name: str) -> mypy_test.User"',
                line=13,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "def (self: mypy_test.User, *, name: str)"',
                line=14,
                column=13,
            ),
        ]
    )
    assert results.ty == snapshot(
        [
            Result(
                type="error",
                message="No argument provided for required parameter `name`",
                line=11,
                column=1,
            ),
            Result(
                type="error",
                message="Argument `n` does not match any known parameter",
                line=11,
                column=6,
            ),
            Result(
                type="information",
                message="Revealed type: `<class 'User'>`",
                line=13,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `(self: User, *, name: str) -> None`",
                line=14,
                column=13,
            ),
        ]
    )



import strawberry

from .type_fruit import Fruit

# Simulate what a third-party decorator (e.g. strawberry-django's filter_type) does:
# re-process a type, creating a new StrawberryObjectDefinition for the same origin class.
strawberry.type(Fruit)


@strawberry.type
class FruitBowl:
    fruit: Fruit



from __future__ import annotations

import contextlib
import json
from collections.abc import AsyncGenerator, Mapping, Sequence
from datetime import timedelta
from io import BytesIO
from typing import Any, Literal

from starlette.requests import Request
from starlette.responses import Response as StarletteResponse
from starlette.testclient import TestClient, WebSocketTestSession
from starlette.websockets import WebSocket

from strawberry.asgi import GraphQL as BaseGraphQLView
from strawberry.http import GraphQLHTTPResponse
from strawberry.http.ides import GraphQL_IDE
from strawberry.schema import Schema
from strawberry.subscriptions import (
    GRAPHQL_TRANSPORT_WS_PROTOCOL,
    GRAPHQL_WS_PROTOCOL,
)
from strawberry.types import ExecutionResult
from tests.http.context import get_context
from tests.views.schema import Query
from tests.websockets.views import OnWSConnectMixin

from .base import (
    JSON,
    DebuggableGraphQLTransportWSHandler,
    DebuggableGraphQLWSHandler,
    HttpClient,
    Message,
    Response,
    ResultOverrideFunction,
    WebSocketClient,
)


class GraphQLView(OnWSConnectMixin, BaseGraphQLView[dict[str, object], object]):
    result_override: ResultOverrideFunction = None
    graphql_transport_ws_handler_class = DebuggableGraphQLTransportWSHandler
    graphql_ws_handler_class = DebuggableGraphQLWSHandler

    async def get_root_value(self, request: WebSocket | Request) -> Query:
        return Query()

    async def get_context(
        self,
        request: Request | WebSocket,
        response: StarletteResponse | WebSocket,
    ) -> dict[str, object]:
        context = await super().get_context(request, response)

        return get_context(context)

    async def process_result(
        self, request: Request, result: ExecutionResult
    ) -> GraphQLHTTPResponse:
        if self.result_override:
            return self.result_override(result)

        return await super().process_result(request, result)


class AsgiHttpClient(HttpClient):
    def __init__(
        self,
        schema: Schema,
        graphql_ide: GraphQL_IDE | None = "graphiql",
        allow_queries_via_get: bool = True,
        keep_alive: bool = False,
        keep_alive_interval: float = 1,
        subscription_protocols: Sequence[str] = (
            GRAPHQL_TRANSPORT_WS_PROTOCOL,
            GRAPHQL_WS_PROTOCOL,
        ),
        connection_init_wait_timeout: timedelta = timedelta(minutes=1),
        result_override: ResultOverrideFunction = None,
        multipart_uploads_enabled: bool = False,
        max_subscriptions_per_connection: int | None = 100,
    ):
        view = GraphQLView(
            schema,
            graphql_ide=graphql_ide,
            allow_queries_via_get=allow_queries_via_get,
            keep_alive=keep_alive,
            keep_alive_interval=keep_alive_interval,
            subscription_protocols=subscription_protocols,
            connection_init_wait_timeout=connection_init_wait_timeout,
            multipart_uploads_enabled=multipart_uploads_enabled,
            max_subscriptions_per_connection=max_subscriptions_per_connection,
        )
        view.result_override = result_override

        self.client = TestClient(view)

    async def _graphql_request(
        self,
        method: Literal["get", "post"],
        query: str | None = None,
        operation_name: str | None = None,
        variables: dict[str, object] | None = None,
        files: dict[str, BytesIO] | None = None,
        headers: dict[str, str] | None = None,
        extensions: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Response:
        body = self._build_body(
            query=query,
            operation_name=operation_name,
            variables=variables,
            files=files,
            method=method,
            extensions=extensions,
        )

        if method == "get":
            kwargs["params"] = body
        elif body:
            if files:
                kwargs["data"] = body
            else:
                kwargs["content"] = json.dumps(body)

        if files is not None:
            kwargs["files"] = files

        response = getattr(self.client, method)(
            "/graphql",
            headers=self._get_headers(method=method, headers=headers, files=files),
            **kwargs,
        )

        return Response(
            status_code=response.status_code,
            data=response.content,
            headers=response.headers,
        )

    async def request(
        self,
        url: str,
        method: Literal["get", "post", "patch", "put", "delete"],
        headers: dict[str, str] | None = None,
    ) -> Response:
        response = getattr(self.client, method)(url, headers=headers)

        return Response(
            status_code=response.status_code,
            data=response.content,
            headers=response.headers,
        )

    async def get(
        self,
        url: str,
        headers: dict[str, str] | None = None,
    ) -> Response:
        return await self.request(url, "get", headers=headers)

    async def post(
        self,
        url: str,
        data: bytes | None = None,
        json: JSON | None = None,
        headers: dict[str, str] | None = None,
    ) -> Response:
        response = self.client.post(url, headers=headers, content=data, json=json)

        return Response(
            status_code=response.status_code,
            data=response.content,
            headers=dict(response.headers),
        )

    @contextlib.asynccontextmanager
    async def ws_connect(
        self,
        url: str,
        *,
        protocols: list[str],
    ) -> AsyncGenerator[WebSocketClient, None]:
        with self.client.websocket_connect(url, protocols) as ws:
            yield AsgiWebSocketClient(ws)


class AsgiWebSocketClient(WebSocketClient):
    def __init__(self, ws: WebSocketTestSession):
        self.ws = ws
        self._closed: bool = False
        self._close_code: int | None = None
        self._close_reason: str | None = None

    async def send_text(self, payload: str) -> None:
        self.ws.send_text(payload)

    async def send_json(self, payload: Mapping[str, object]) -> None:
        self.ws.send_json(payload)

    async def send_bytes(self, payload: bytes) -> None:
        self.ws.send_bytes(payload)

    async def receive(self, timeout: float | None = None) -> Message:
        if self._closed:
            # if close was received via exception, fake it so that recv works
            return Message(
                type="websocket.close", data=self._close_code, extra=self._close_reason
            )
        m = self.ws.receive()
        if m["type"] == "websocket.close":
            self._closed = True
            self._close_code = m["code"]
            self._close_reason = m["reason"]
            return Message(type=m["type"], data=m["code"], extra=m["reason"])
        if m["type"] == "websocket.send":
            return Message(type=m["type"], data=m["text"])
        return Message(type=m["type"], data=m["data"], extra=m["extra"])

    async def receive_json(self, timeout: float | None = None) -> Any:
        m = self.ws.receive()
        assert m["type"] == "websocket.send"
        assert "text" in m
        return json.loads(m["text"])

    async def close(self) -> None:
        self.ws.close()
        self._closed = True

    @property
    def accepted_subprotocol(self) -> str | None:
        return self.ws.accepted_subprotocol

    @property
    def closed(self) -> bool:
        return self._closed

    @property
    def close_code(self) -> int:
        assert self._close_code is not None
        return self._close_code

    @property
    def close_reason(self) -> str | None:
        return self._close_reason



from django.urls import path

from strawberry.django.views import GraphQLView as BaseGraphQLView
from tests.views.schema import Query, schema


class GraphQLView(BaseGraphQLView):
    def get_root_value(self, request) -> Query:
        return Query()


urlpatterns = [
    path("graphql/", GraphQLView.as_view(schema=schema)),
]



import strawberry


def test_base_context():
    from strawberry.litestar import BaseContext

    base_context = BaseContext()
    assert base_context.request is None


def test_with_class_context_getter():
    from litestar import Litestar
    from litestar.di import Provide
    from litestar.testing import TestClient

    from strawberry.litestar import BaseContext, make_graphql_controller

    @strawberry.type
    class Query:
        @strawberry.field
        def abc(self, info: strawberry.Info) -> str:
            assert isinstance(info.context, CustomContext)
            assert info.context.request is not None
            assert info.context.strawberry == "rocks"
            return "abc"

    class CustomContext(BaseContext):
        strawberry: str

    def custom_context_dependency() -> CustomContext:
        return CustomContext(strawberry="rocks")

    async def get_context(custom_context_dependency: CustomContext):
        return custom_context_dependency

    schema = strawberry.Schema(query=Query)
    graphql_controller = make_graphql_controller(
        path="/graphql", schema=schema, context_getter=get_context
    )
    app = Litestar(
        route_handlers=[graphql_controller],
        dependencies={
            "custom_context_dependency": Provide(
                custom_context_dependency, sync_to_thread=True
            )
        },
    )

    test_client = TestClient(app)
    response = test_client.post("/graphql", json={"query": "{ abc }"})

    assert response.status_code == 200
    assert response.json() == {"data": {"abc": "abc"}}


def test_with_dict_context_getter():
    from litestar import Litestar
    from litestar.di import Provide
    from litestar.testing import TestClient

    from strawberry.litestar import make_graphql_controller

    @strawberry.type
    class Query:
        @strawberry.field
        def abc(self, info: strawberry.Info) -> str:
            assert isinstance(info.context, dict)
            assert info.context.get("request") is not None
            assert info.context.get("strawberry") == "rocks"
            return "abc"

    def custom_context_dependency() -> str:
        return "rocks"

    async def get_context(custom_context_dependency: str) -> dict[str, str]:
        return {"strawberry": custom_context_dependency}

    schema = strawberry.Schema(query=Query)
    graphql_controller = make_graphql_controller(
        path="/graphql", schema=schema, context_getter=get_context
    )
    app = Litestar(
        route_handlers=[graphql_controller],
        dependencies={
            "custom_context_dependency": Provide(
                custom_context_dependency, sync_to_thread=True
            )
        },
    )
    test_client = TestClient(app)
    response = test_client.post("/graphql", json={"query": "{ abc }"})

    assert response.status_code == 200
    assert response.json() == {"data": {"abc": "abc"}}


def test_without_context_getter():
    from litestar import Litestar
    from litestar.testing import TestClient

    from strawberry.litestar import make_graphql_controller

    @strawberry.type
    class Query:
        @strawberry.field
        def abc(self, info: strawberry.Info) -> str:
            assert isinstance(info.context, dict)
            assert info.context.get("request") is not None
            assert info.context.get("strawberry") is None
            return "abc"

    schema = strawberry.Schema(query=Query)
    graphql_controller = make_graphql_controller(
        path="/graphql", schema=schema, context_getter=None
    )
    app = Litestar(route_handlers=[graphql_controller])
    test_client = TestClient(app)
    response = test_client.post("/graphql", json={"query": "{ abc }"})

    assert response.status_code == 200
    assert response.json() == {"data": {"abc": "abc"}}


def test_with_invalid_context_getter():
    from litestar import Litestar
    from litestar.di import Provide
    from litestar.testing import TestClient

    from strawberry.litestar import make_graphql_controller

    @strawberry.type
    class Query:
        @strawberry.field
        def abc(self, info: strawberry.Info) -> str:
            assert info.context.get("request") is not None
            assert info.context.get("strawberry") is None
            return "abc"

    def custom_context_dependency() -> str:
        return "rocks"

    async def get_context(custom_context_dependency: str) -> str:
        return custom_context_dependency

    schema = strawberry.Schema(query=Query)
    graphql_controller = make_graphql_controller(
        path="/graphql", schema=schema, context_getter=get_context
    )
    app = Litestar(
        route_handlers=[graphql_controller],
        dependencies={
            "custom_context_dependency": Provide(
                custom_context_dependency, sync_to_thread=True
            )
        },
    )
    test_client = TestClient(app, raise_server_exceptions=True)
    # TODO: test exception message
    # assert starlite.exceptions.http_exceptions.InternalServerException is raised
    # with pytest.raises(
    #     InternalServerException,
    #     r"A dependency failed validation for POST .*"
    # ),
    # ):
    response = test_client.post("/graphql", json={"query": "{ abc }"})
    assert response.status_code == 500
    assert response.json()["detail"] == "Internal Server Error"


def test_custom_context():
    from litestar.testing import TestClient

    from tests.litestar.app import create_app

    @strawberry.type
    class Query:
        @strawberry.field
        def custom_context_value(self, info: strawberry.Info) -> str:
            return info.context["custom_value"]

    schema = strawberry.Schema(query=Query)
    app = create_app(schema=schema)

    test_client = TestClient(app)
    response = test_client.post("/graphql", json={"query": "{ customContextValue }"})

    assert response.status_code == 200
    assert response.json() == {"data": {"customContextValue": "Hi!"}}


def test_can_set_background_task():
    from litestar.testing import TestClient

    from tests.litestar.app import create_app

    task_complete = False

    async def task():
        nonlocal task_complete
        task_complete = True

    @strawberry.type
    class Query:
        @strawberry.field
        def something(self, info: strawberry.Info) -> str:
            response = info.context["response"]
            response.background.tasks.append(task)
            return "foo"

    schema = strawberry.Schema(query=Query)
    app = create_app(schema=schema)

    test_client = TestClient(app)
    response = test_client.post("/graphql", json={"query": "{ something }"})

    assert response.json() == {"data": {"something": "foo"}}
    assert task_complete



from __future__ import annotations

import textwrap
from uuid import UUID

import strawberry
from strawberry.field_extensions import InputMutationExtension


@strawberry.type
class Query:
    @strawberry.field
    async def hello(self) -> str:
        return "hi"


@strawberry.type
class Mutation:
    @strawberry.mutation(extensions=[InputMutationExtension()])
    async def buggy(self, some_id: UUID) -> None:
        del some_id


def test_schema():
    schema = strawberry.Schema(query=Query, mutation=Mutation)

    expected_schema = '''
    input BuggyInput {
      someId: UUID!
    }

    type Mutation {
      buggy(
        """Input data for `buggy` mutation"""
        input: BuggyInput!
      ): Void
    }

    type Query {
      hello: String!
    }

    scalar UUID

    """Represents NULL values"""
    scalar Void
    '''

    assert textwrap.dedent(expected_schema).strip() == str(schema).strip()



import contextlib
import json
import urllib.parse
from collections.abc import AsyncGenerator, Sequence
from datetime import timedelta
from io import BytesIO
from typing import Any, Literal

from quart import Quart
from quart import Request as QuartRequest
from quart import Response as QuartResponse
from quart import Websocket as QuartWebsocket
from quart.datastructures import FileStorage
from starlette.testclient import TestClient
from starlette.types import Receive, Scope, Send

from strawberry.http import GraphQLHTTPResponse
from strawberry.http.ides import GraphQL_IDE
from strawberry.quart.views import GraphQLView as BaseGraphQLView
from strawberry.schema import Schema
from strawberry.subscriptions import (
    GRAPHQL_TRANSPORT_WS_PROTOCOL,
    GRAPHQL_WS_PROTOCOL,
)
from strawberry.types import ExecutionResult
from tests.http.context import get_context
from tests.views.schema import Query
from tests.websockets.views import OnWSConnectMixin

from .asgi import AsgiWebSocketClient
from .base import (
    JSON,
    DebuggableGraphQLTransportWSHandler,
    DebuggableGraphQLWSHandler,
    HttpClient,
    Response,
    ResultOverrideFunction,
    WebSocketClient,
)


class GraphQLView(OnWSConnectMixin, BaseGraphQLView[dict[str, object], object]):
    methods = ["GET", "POST", "PUT", "PATCH", "DELETE", "HEAD"]
    result_override: ResultOverrideFunction = None
    graphql_transport_ws_handler_class = DebuggableGraphQLTransportWSHandler
    graphql_ws_handler_class = DebuggableGraphQLWSHandler

    def __init__(self, *args: Any, **kwargs: Any):
        self.result_override = kwargs.pop("result_override", None)
        super().__init__(*args, **kwargs)

    async def get_root_value(self, request: QuartRequest | QuartWebsocket) -> Query:
        await super().get_root_value(request)  # for coverage
        return Query()

    async def get_context(
        self, request: QuartRequest | QuartWebsocket, response: QuartResponse
    ) -> dict[str, object]:
        context = await super().get_context(request, response)

        return get_context(context)

    async def process_result(
        self, request: QuartRequest, result: ExecutionResult
    ) -> GraphQLHTTPResponse:
        if self.result_override:
            return self.result_override(result)

        return await super().process_result(request, result)


class QuartAsgiAppAdapter:
    def __init__(self, app: Quart):
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["asgi"] = scope.get("asgi", {})

        # Our WebSocket tests depend on WebSocket close reasons.
        # Quart only sends close reason if the ASGI spec version in the scope is => 2.3
        # https://github.com/pallets/quart/blob/b5593ca4c8c657564cdf2d35c9f0298fce63636b/src/quart/asgi.py#L347-L348
        scope["asgi"]["spec_version"] = "2.3"

        await self.app(scope, receive, send)  # type: ignore


class QuartHttpClient(HttpClient):
    def __init__(
        self,
        schema: Schema,
        graphql_ide: GraphQL_IDE | None = "graphiql",
        allow_queries_via_get: bool = True,
        keep_alive: bool = False,
        keep_alive_interval: float = 1,
        subscription_protocols: Sequence[str] = (
            GRAPHQL_TRANSPORT_WS_PROTOCOL,
            GRAPHQL_WS_PROTOCOL,
        ),
        connection_init_wait_timeout: timedelta = timedelta(minutes=1),
        result_override: ResultOverrideFunction = None,
        multipart_uploads_enabled: bool = False,
        max_subscriptions_per_connection: int | None = 100,
    ):
        self.app = Quart(__name__)
        self.app.debug = True

        view = GraphQLView.as_view(
            "graphql_view",
            schema=schema,
            graphql_ide=graphql_ide,
            allow_queries_via_get=allow_queries_via_get,
            result_override=result_override,
            keep_alive=keep_alive,
            keep_alive_interval=keep_alive_interval,
            subscription_protocols=subscription_protocols,
            connection_init_wait_timeout=connection_init_wait_timeout,
            multipart_uploads_enabled=multipart_uploads_enabled,
            max_subscriptions_per_connection=max_subscriptions_per_connection,
        )

        self.app.add_url_rule(
            "/graphql",
            view_func=view,
        )

        self.app.add_url_rule(
            "/graphql",
            view_func=view,
            methods=["GET"],
            websocket=True,
        )

        self.client = TestClient(QuartAsgiAppAdapter(self.app))

    async def _graphql_request(
        self,
        method: Literal["get", "post"],
        query: str | None = None,
        operation_name: str | None = None,
        variables: dict[str, object] | None = None,
        files: dict[str, BytesIO] | None = None,
        headers: dict[str, str] | None = None,
        extensions: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Response:
        body = self._build_body(
            query=query,
            operation_name=operation_name,
            variables=variables,
            files=files,
            method=method,
            extensions=extensions,
        )

        url = "/graphql"

        if method == "get":
            body_encoded = urllib.parse.urlencode(body or {})
            url = f"{url}?{body_encoded}"
        elif body:
            if files:
                kwargs["form"] = body
                kwargs["files"] = {
                    k: FileStorage(v, filename=k) for k, v in files.items()
                }
            else:
                kwargs["data"] = json.dumps(body)

        headers = self._get_headers(method=method, headers=headers, files=files)

        return await self.request(url, method, headers=headers, **kwargs)

    async def request(
        self,
        url: str,
        method: Literal["head", "get", "post", "patch", "put", "delete"],
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> Response:
        async with self.app.test_app() as test_app, self.app.app_context():
            client = test_app.test_client()
            response = await getattr(client, method)(url, headers=headers, **kwargs)

        return Response(
            status_code=response.status_code,
            data=(await response.data),
            headers=response.headers,
        )

    async def get(
        self,
        url: str,
        headers: dict[str, str] | None = None,
    ) -> Response:
        return await self.request(url, "get", headers=headers)

    async def post(
        self,
        url: str,
        data: bytes | None = None,
        json: JSON | None = None,
        headers: dict[str, str] | None = None,
    ) -> Response:
        kwargs = {"headers": headers, "data": data, "json": json}
        return await self.request(
            url, "post", **{k: v for k, v in kwargs.items() if v is not None}
        )

    @contextlib.asynccontextmanager
    async def ws_connect(
        self,
        url: str,
        *,
        protocols: list[str],
    ) -> AsyncGenerator[WebSocketClient, None]:
        with self.client.websocket_connect(url, protocols) as ws:
            yield AsgiWebSocketClient(ws)



import strawberry


def test_field_descriptions():
    description = "this description is super cool"

    field = strawberry.field(description=description)

    assert field.description == description



from datetime import date, datetime, timedelta, timezone
from decimal import Decimal
from textwrap import dedent
from uuid import UUID

import pytest

import strawberry
from strawberry import scalar
from strawberry.exceptions import ScalarAlreadyRegisteredError
from strawberry.scalars import JSON, Base16, Base32, Base64


def test_void_function():
    NoneType = type(None)

    @strawberry.type
    class Query:
        @strawberry.field
        def void_ret(self) -> None:
            return

        @strawberry.field
        def void_ret_crash(self) -> NoneType:
            return 1

        @strawberry.field
        def void_arg(self, x: None) -> None:
            return

    schema = strawberry.Schema(query=Query)

    assert (
        str(schema)
        == dedent(
            '''
      type Query {
        voidRet: Void
        voidRetCrash: Void
        voidArg(x: Void): Void
      }

      """Represents NULL values"""
      scalar Void
    '''
        ).strip()
    )

    result = schema.execute_sync("query { voidRet }")
    assert not result.errors
    assert result.data == {
        "voidRet": None,
    }

    result = schema.execute_sync("query { voidArg (x: null) }")
    assert not result.errors
    assert result.data == {
        "voidArg": None,
    }

    result = schema.execute_sync("query { voidArg (x: 1) }")
    assert result.errors

    result = schema.execute_sync("query { voidRetCrash }")
    assert result.errors


def test_uuid_field_string_value():
    @strawberry.type
    class Query:
        unique_id: UUID

    schema = strawberry.Schema(query=Query)

    assert (
        str(schema)
        == dedent(
            """
      type Query {
        uniqueId: UUID!
      }

      scalar UUID
    """
        ).strip()
    )

    result = schema.execute_sync(
        "query { uniqueId }",
        root_value=Query(
            unique_id="e350746c-33b6-4469-86b0-5f16e1e12232",
        ),
    )
    assert not result.errors
    assert result.data == {
        "uniqueId": "e350746c-33b6-4469-86b0-5f16e1e12232",
    }


def test_uuid_field_uuid_value():
    @strawberry.type
    class Query:
        unique_id: UUID

    schema = strawberry.Schema(query=Query)

    assert (
        str(schema)
        == dedent(
            """
      type Query {
        uniqueId: UUID!
      }

      scalar UUID
    """
        ).strip()
    )

    result = schema.execute_sync(
        "query { uniqueId }",
        root_value=Query(
            unique_id=UUID("e350746c-33b6-4469-86b0-5f16e1e12232"),
        ),
    )
    assert not result.errors
    assert result.data == {
        "uniqueId": "e350746c-33b6-4469-86b0-5f16e1e12232",
    }


def test_uuid_input():
    @strawberry.type
    class Query:
        ok: bool

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def uuid_input(self, input_id: UUID) -> str:
            assert isinstance(input_id, UUID)
            return str(input_id)

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    result = schema.execute_sync(
        """
        mutation {
            uuidInput(inputId: "e350746c-33b6-4469-86b0-5f16e1e12232")
        }
    """
    )

    assert not result.errors
    assert result.data == {
        "uuidInput": "e350746c-33b6-4469-86b0-5f16e1e12232",
    }


def test_json():
    @strawberry.type
    class Query:
        @strawberry.field
        def echo_json(data: JSON) -> JSON:
            return data

        @strawberry.field
        def echo_json_nullable(data: JSON | None) -> JSON | None:
            return data

    schema = strawberry.Schema(query=Query)

    expected_schema = dedent(
        '''
        """
        The `JSON` scalar type represents JSON values as specified by [ECMA-404](https://ecma-international.org/wp-content/uploads/ECMA-404_2nd_edition_december_2017.pdf).
        """
        scalar JSON @specifiedBy(url: "https://ecma-international.org/wp-content/uploads/ECMA-404_2nd_edition_december_2017.pdf")

        type Query {
          echoJson(data: JSON!): JSON!
          echoJsonNullable(data: JSON): JSON
        }
        '''
    ).strip()

    assert str(schema) == expected_schema

    result = schema.execute_sync(
        """
        query {
            echoJson(data: {hello: {a: 1}, someNumbers: [1, 2, 3], null: null})
            echoJsonNullable(data: {hello: {a: 1}, someNumbers: [1, 2, 3], null: null})
        }
    """
    )

    assert not result.errors
    assert result.data == {
        "echoJson": {"hello": {"a": 1}, "someNumbers": [1, 2, 3], "null": None},
        "echoJsonNullable": {"hello": {"a": 1}, "someNumbers": [1, 2, 3], "null": None},
    }

    result = schema.execute_sync(
        """
        query {
            echoJson(data: null)
        }
    """
    )
    assert result.errors  # echoJson is not-null null

    result = schema.execute_sync(
        """
        query {
            echoJsonNullable(data: null)
        }
    """
    )
    assert not result.errors
    assert result.data == {
        "echoJsonNullable": None,
    }


def test_base16():
    @strawberry.type
    class Query:
        @strawberry.field
        def base16_encode(data: str) -> Base16:
            return bytes(data, "utf-8")

        @strawberry.field
        def base16_decode(data: Base16) -> str:
            return data.decode("utf-8")

        @strawberry.field
        def base32_encode(data: str) -> Base32:
            return bytes(data, "utf-8")

        @strawberry.field
        def base32_decode(data: Base32) -> str:
            return data.decode("utf-8")

        @strawberry.field
        def base64_encode(data: str) -> Base64:
            return bytes(data, "utf-8")

        @strawberry.field
        def base64_decode(data: Base64) -> str:
            return data.decode("utf-8")

    schema = strawberry.Schema(query=Query)

    assert (
        str(schema)
        == dedent(
            '''
        """Represents binary data as Base16-encoded (hexadecimal) strings."""
        scalar Base16 @specifiedBy(url: "https://datatracker.ietf.org/doc/html/rfc4648.html#section-8")

        """
        Represents binary data as Base32-encoded strings, using the standard alphabet.
        """
        scalar Base32 @specifiedBy(url: "https://datatracker.ietf.org/doc/html/rfc4648.html#section-6")

        """
        Represents binary data as Base64-encoded strings, using the standard alphabet.
        """
        scalar Base64 @specifiedBy(url: "https://datatracker.ietf.org/doc/html/rfc4648.html#section-4")

        type Query {
          base16Encode(data: String!): Base16!
          base16Decode(data: Base16!): String!
          base32Encode(data: String!): Base32!
          base32Decode(data: Base32!): String!
          base64Encode(data: String!): Base64!
          base64Decode(data: Base64!): String!
        }
    '''
        ).strip()
    )

    result = schema.execute_sync(
        """
        query {
            base16Encode(data: "Hello")
            base16Decode(data: "48656c6C6f")  # < Mix lowercase and uppercase
            base32Encode(data: "Hello")
            base32Decode(data: "JBSWY3dp")  # < Mix lowercase and uppercase
            base64Encode(data: "Hello")
            base64Decode(data: "SGVsbG8=")
        }
    """
    )

    assert not result.errors
    assert result.data == {
        "base16Encode": "48656C6C6F",
        "base16Decode": "Hello",
        "base32Encode": "JBSWY3DP",
        "base32Decode": "Hello",
        "base64Encode": "SGVsbG8=",
        "base64Decode": "Hello",
    }


def test_override_built_in_scalars():
    EpochDateTime = strawberry.scalar(
        datetime,
        serialize=lambda value: int(value.timestamp()),
        parse_value=lambda value: datetime.fromtimestamp(int(value), timezone.utc),
    )

    @strawberry.type
    class Query:
        @strawberry.field
        def current_time(self) -> datetime:
            return datetime(2021, 8, 11, 12, 0, tzinfo=timezone.utc)

        @strawberry.field
        def isoformat(self, input_datetime: datetime) -> str:
            return input_datetime.isoformat()

    schema = strawberry.Schema(
        Query,
        scalar_overrides={
            datetime: EpochDateTime,
        },
    )

    result = schema.execute_sync(
        """
        {
            currentTime
            isoformat(inputDatetime: 1628683200)
        }
        """
    )

    assert not result.errors
    assert result.data["currentTime"] == 1628683200
    assert result.data["isoformat"] == "2021-08-11T12:00:00+00:00"


def test_override_unknown_scalars():
    Duration = strawberry.scalar(
        timedelta,
        name="Duration",
        serialize=timedelta.total_seconds,
        parse_value=lambda s: timedelta(seconds=s),
    )

    @strawberry.type
    class Query:
        @strawberry.field
        def duration(self, value: timedelta) -> timedelta:
            return value

    schema = strawberry.Schema(Query, scalar_overrides={timedelta: Duration})

    result = schema.execute_sync("{ duration(value: 10) }")

    assert not result.errors
    assert result.data == {"duration": 10}


def test_decimal():
    @strawberry.type
    class Query:
        @strawberry.field
        def decimal(value: Decimal) -> Decimal:
            return value

    schema = strawberry.Schema(query=Query)

    result = schema.execute_sync(
        """
        query {
            floatDecimal: decimal(value: 3.14)
            floatDecimal2: decimal(value: 3.14509999)
            floatDecimal3: decimal(value: 0.000001)
            stringDecimal: decimal(value: "3.14")
            stringDecimal2: decimal(value: "3.1499999991")
        }
    """
    )

    assert not result.errors
    assert result.data == {
        "floatDecimal": "3.14",
        "floatDecimal2": "3.14509999",
        "floatDecimal3": "0.000001",
        "stringDecimal": "3.14",
        "stringDecimal2": "3.1499999991",
    }


@pytest.mark.raises_strawberry_exception(
    ScalarAlreadyRegisteredError,
    match="Scalar `MyCustomScalar` has already been registered",
)
def test_duplicate_scalars_raises_exception():
    MyCustomScalar = strawberry.scalar(
        str,
        name="MyCustomScalar",
    )

    MyCustomScalar2 = strawberry.scalar(
        int,
        name="MyCustomScalar",
    )

    @strawberry.type
    class Query:
        scalar_1: MyCustomScalar
        scalar_2: MyCustomScalar2

    strawberry.Schema(Query)


@pytest.mark.raises_strawberry_exception(
    ScalarAlreadyRegisteredError,
    match="Scalar `MyCustomScalar` has already been registered",
)
def test_duplicate_scalars_raises_exception_using_alias():
    MyCustomScalar = scalar(
        str,
        name="MyCustomScalar",
    )

    MyCustomScalar2 = scalar(
        int,
        name="MyCustomScalar",
    )

    @strawberry.type
    class Query:
        scalar_1: MyCustomScalar
        scalar_2: MyCustomScalar2

    strawberry.Schema(Query)


def test_optional_scalar_with_or_operator():
    """Check `|` operator support with an optional scalar."""

    @strawberry.type
    class Query:
        date: date | None

    schema = strawberry.Schema(query=Query)

    query = "{ date }"

    result = schema.execute_sync(query, root_value=Query(date=None))
    assert not result.errors
    assert result.data["date"] is None

    result = schema.execute_sync(query, root_value=Query(date=date(2020, 1, 1)))
    assert not result.errors
    assert result.data["date"] == "2020-01-01"


def test_scalar_map_with_newtype():
    """Test scalar_map configuration with NewType."""
    from typing import NewType

    from strawberry.schema.config import StrawberryConfig

    MyString = NewType("MyString", str)

    @strawberry.type
    class Query:
        @strawberry.field
        def my_string(self) -> MyString:
            return MyString("hello")

        @strawberry.field
        def process_string(self, value: MyString) -> str:
            return f"processed: {value}"

    schema = strawberry.Schema(
        query=Query,
        config=StrawberryConfig(
            scalar_map={
                MyString: strawberry.scalar(
                    name="MyString",
                    description="A custom string scalar",
                    serialize=lambda v: v.upper(),
                    parse_value=lambda v: v.lower(),
                )
            }
        ),
    )

    expected_schema = dedent(
        '''
        """A custom string scalar"""
        scalar MyString

        type Query {
          myString: MyString!
          processString(value: MyString!): String!
        }
        '''
    ).strip()

    assert str(schema) == expected_schema

    result = schema.execute_sync("{ myString }")
    assert not result.errors
    assert result.data == {"myString": "HELLO"}

    result = schema.execute_sync('{ processString(value: "WORLD") }')
    assert not result.errors
    assert result.data == {"processString": "processed: world"}


def test_scalar_map_override_builtin():
    """Test scalar_map can override built-in scalars."""
    from strawberry.schema.config import StrawberryConfig

    @strawberry.type
    class Query:
        @strawberry.field
        def current_time(self) -> datetime:
            return datetime(2021, 8, 11, 12, 0, tzinfo=timezone.utc)

        @strawberry.field
        def parse_time(self, value: datetime) -> str:
            return value.isoformat()

    schema = strawberry.Schema(
        query=Query,
        config=StrawberryConfig(
            scalar_map={
                datetime: strawberry.scalar(
                    name="DateTime",
                    serialize=lambda v: int(v.timestamp()),
                    parse_value=lambda v: datetime.fromtimestamp(int(v), timezone.utc),
                )
            }
        ),
    )

    result = schema.execute_sync("{ currentTime parseTime(value: 1628683200) }")
    assert not result.errors
    assert result.data["currentTime"] == 1628683200
    assert result.data["parseTime"] == "2021-08-11T12:00:00+00:00"


def test_scalar_map_with_specified_by_url():
    """Test scalar_map with specified_by_url."""
    from typing import NewType

    from strawberry.schema.config import StrawberryConfig

    Email = NewType("Email", str)

    @strawberry.type
    class Query:
        @strawberry.field
        def email(self) -> Email:
            return Email("test@example.com")

    schema = strawberry.Schema(
        query=Query,
        config=StrawberryConfig(
            scalar_map={
                Email: strawberry.scalar(
                    name="Email",
                    description="An email address",
                    specified_by_url="https://html.spec.whatwg.org/multipage/input.html#valid-e-mail-address",
                )
            }
        ),
    )

    expected = 'scalar Email @specifiedBy(url: "https://html.spec.whatwg.org/multipage/input.html#valid-e-mail-address")'
    assert expected in str(schema)


def test_scalar_map_combined_with_scalar_overrides():
    """Test that scalar_map and scalar_overrides work together."""
    import warnings
    from typing import NewType

    from strawberry.schema.config import StrawberryConfig

    MyInt = NewType("MyInt", int)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        MyFloat = strawberry.scalar(
            float,
            name="MyFloat",
            serialize=lambda v: round(v, 2),
        )

    @strawberry.type
    class Query:
        my_int: MyInt
        my_float: MyFloat

    schema = strawberry.Schema(
        query=Query,
        config=StrawberryConfig(
            scalar_map={
                MyInt: strawberry.scalar(
                    name="MyInt",
                    serialize=lambda v: v * 2,
                )
            }
        ),
        scalar_overrides={
            float: MyFloat,
        },
    )

    result = schema.execute_sync(
        "{ myInt myFloat }", root_value=Query(my_int=5, my_float=3.14159)
    )
    assert not result.errors
    assert result.data == {"myInt": 10, "myFloat": 3.14}


def test_scalar_definition_direct_creation():
    """Test creating ScalarDefinition directly via scalar()."""
    definition = strawberry.scalar(
        name="TestScalar",
        description="A test scalar",
        serialize=str,
        parse_value=int,
    )

    assert definition.name == "TestScalar"
    assert definition.description == "A test scalar"
    assert definition.serialize is not None
    assert definition.parse_value is not None


def test_builtin_scalars_are_newtypes():
    """Test that built-in scalars are proper NewType instances."""
    from strawberry.scalars import ID, JSON, Base16, Base32, Base64

    assert ID.__supertype__ is str
    assert JSON.__supertype__ is object
    assert Base16.__supertype__ is bytes
    assert Base32.__supertype__ is bytes
    assert Base64.__supertype__ is bytes
